# Fluster - testing framework for decoders conformance
# Copyright (C) 2021, Fluendo, S.A.
#  Author: Manuel Jimeno Martínez <mjimeno@fluendo.com>, Fluendo, S.A.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=import-outside-toplevel
from functools import lru_cache
from fluster.codec import Codec, OutputFormat
from fluster.decoder import Decoder, register_decoder

try:
    from fluster_chromium import main  # type: ignore

    HAS_FLUSTER_CHROMIUM = True
except ImportError:
    HAS_FLUSTER_CHROMIUM = False


@register_decoder
class ChromiumH264(Decoder):
    """H264 class for Chromium decoder"""

    provider = "Chromium"
    codec = Codec.H264
    description = "This is an implementation for the Chromium decoder"

    def __init__(self) -> None:
        super().__init__()
        if not self.name:
            self.name = f"{self.provider}-{self.codec.value}"
        self.description = f"{self.provider} {self.codec.value} decoder for Chromium"

    def decode(
        self,
        input_filepath: str,
        output_filepath: str,
        output_format: OutputFormat,
        timeout: int,
        verbose: bool,
        keep_files: bool,
    ) -> str:
        # pylint: disable=unused-argument
        return str(main(input_filepath))

    @lru_cache(maxsize=128)
    def check(self, verbose: bool) -> bool:
        if verbose and not HAS_FLUSTER_CHROMIUM:
            print("Decoder is not available, module fluster-chromium not installed")
        return HAS_FLUSTER_CHROMIUM



# Fluster - testing framework for decoders conformance
# Copyright (C) 2020, Fluendo, S.A.
#  Author: Pablo Marcos Oltra <pmarcos@fluendo.com>, Fluendo, S.A.
#  Author: Andoni Morales Alastruey <amorales@fluendo.com>, Fluendo, S.A.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <https://www.gnu.org/licenses/>.

from fluster.codec import Codec, OutputFormat
from fluster.decoder import Decoder, register_decoder
from fluster.utils import file_checksum, run_command


class VPXDecoder(Decoder):
    """Generic class for VPX reference decoder"""

    name = ""
    description = ""
    binary = "vpxdec"
    codec = Codec.NONE

    def __init__(self) -> None:
        super().__init__()
        self.name = f"libvpx-{self.codec.value}"
        self.description = f"{self.codec.value} reference decoder"

    def decode(
        self,
        input_filepath: str,
        output_filepath: str,
        output_format: OutputFormat,
        timeout: int,
        verbose: bool,
        keep_files: bool,
    ) -> str:
        """Decodes input_filepath in output_filepath"""
        # pylint: disable=unused-argument
        fmt = None
        if output_format == OutputFormat.YUV420P:
            fmt = "--i420"
        else:
            fmt = "--rawvideo"

        run_command(
            [self.binary, fmt, input_filepath, "-o", output_filepath],
            timeout=timeout,
            verbose=verbose,
        )
        return file_checksum(output_filepath)


@register_decoder
class VP8Decoder(VPXDecoder):
    """VP8 reference decoder implementation"""

    codec = Codec.VP8


@register_decoder
class VP9Decoder(VPXDecoder):
    """VP9 reference decoder implementation"""

    codec = Codec.VP9



# Fluster - testing framework for decoders conformance
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <https://www.gnu.org/licenses/>.

import os
import argparse
import sys
import multiprocessing
import urllib.request
import json
import re

# pylint: disable=wrong-import-position
sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
from fluster.test_suite import TestSuite, TestVector
from fluster.codec import Codec, OutputFormat
from fluster import utils
from fluster.decoders import av1_aom

# pylint: enable=wrong-import-position

DOWNLOAD_URL = "https://storage.googleapis.com/chromiumos-test-assets-public/tast/cros/video/test_vectors/av1"

TESTS_8BPP = (
    # 8 bit
    "00000527_20210205.ivf",
    "00000535_20210205.ivf",
    "00000548_20201006.ivf",
    "48_delayed_20201006.ivf",
    "av1-1-b8-02-allintra_20201006.ivf",
    "av1-1-b8-03-sizeup_20201006.ivf",
    "av1-1-b8-23-film_grain-50_20201006.ivf",
    "ccvb_film_grain_20201006.ivf",
    "frames_refs_short_signaling_20201006.ivf",
    "non_uniform_tiling_20201006.ivf",
    "test-25fps-192x288-only-tile-cols-is-power-of-2_20210111.ivf",
    "test-25fps-192x288-only-tile-rows-is-power-of-2_20210111.ivf",
    "test-25fps-192x288-tile-rows-3-tile-cols-3_20210111.ivf",
)

TESTS_10BPP = (
    #10 bit
    "00000671_20210310.ivf",
    "00000672_20210310.ivf",
    "00000673_20210310.ivf",
    "00000674_20210310.ivf",
    "00000675_20210310.ivf",
    "00000716_20210310.ivf",
    "00000717_20210310.ivf",
    "00000718_20210310.ivf",
    "00000719_20210310.ivf",
    "00000720_20210310.ivf",
    "00000761_20210310.ivf",
    "00000762_20210310.ivf",
    "00000763_20210310.ivf",
    "00000764_20210310.ivf",
    "00000765_20210310.ivf",
    "av1-1-b10-00-quantizer-00_20210310.ivf",
    "av1-1-b10-00-quantizer-10_20210310.ivf",
    "av1-1-b10-00-quantizer-20_20210310.ivf",
    "av1-1-b10-00-quantizer-30_20210310.ivf",
    "av1-1-b10-00-quantizer-40_20210310.ivf",
    "av1-1-b10-00-quantizer-50_20210310.ivf",
    "av1-1-b10-00-quantizer-60_20210310.ivf",
    "av1-1-b10-23-film_grain-50_20210310.ivf"
)

class ChromiumAV1Generator:
    """Generates a test suite from the conformance bitstreams used in tast tests for Chromium"""

    def __init__(
        self,
        name: str,
        suite_name: str,
        codec: Codec,
        description: str,
        bpp: int,
    ):
        self.name = name
        self.suite_name = suite_name
        self.codec = codec
        self.description = description
        self.decoder = av1_aom.AV1AOMDecoder()
        self.bpp = bpp


    def generate(self, download, jobs):
        """Generates the test suite and saves it to a file"""
        output_filepath = os.path.join(self.suite_name + ".json")
        test_suite = TestSuite(
            output_filepath,
            "resources",
            self.suite_name,
            self.codec,
            self.description,
            dict(),
        )

        print(f"Download list of bitstreams from {DOWNLOAD_URL}")

        if self.bpp == 10:
            TESTS = TESTS_10BPP
        elif self.bpp == 8:
            TESTS = TESTS_8BPP
        else:
            return

        for test in TESTS:
            file_url = f"{DOWNLOAD_URL}/{test}"
            name = re.sub("_[\d]*", "", test)

            test_vector = TestVector(
                name, file_url, "__skip__", test, OutputFormat.YUV420P, "")

            test_suite.test_vectors[name] = test_vector

        if download:
            test_suite.download(
                jobs=jobs,
                out_dir=test_suite.resources_dir,
                verify=False,
                extract_all=True,
                keep_file=True,
            )

        for test_vector in test_suite.test_vectors.values():
            dest_dir = os.path.join(
                test_suite.resources_dir, test_suite.name, test_vector.name
            )
            dest_path = os.path.join(
                dest_dir, os.path.basename(test_vector.source))
            test_vector.input_file = dest_path.replace(
                os.path.join(
                    test_suite.resources_dir, test_suite.name, test_vector.name
                )
                + os.sep,
                "",
            )

            if not test_vector.input_file:
                raise Exception(f"Bitstream file not found in {dest_dir}")
            test_vector.source_checksum = utils.file_checksum(dest_path)
            out420 = f"{dest_path}.i420"
            # Run the libaom av1 decoder to get the checksum as the .md5 in the JSONs are per-frame
            test_vector.result = self.decoder.decode(
                dest_path, out420, test_vector.output_format, 30, False)
            os.remove(out420)

        test_suite.to_json_file(output_filepath)
        print("Generate new test suite: " + test_suite.name + ".json")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--skip-download",
        help="skip extracting tarball",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-j",
        "--jobs",
        help="number of parallel jobs to use. 2x logical cores by default",
        type=int,
        default=2 * multiprocessing.cpu_count(),
    )
    args = parser.parse_args()
    generator = ChromiumAV1Generator(
        "chromium-AV1-8bit",
        "CHROMIUM-8bit-AV1-TEST-VECTORS",
        Codec.AV1,
        "AV1 Test Vector Catalogue from https://source.chromium.org/chromiumos/chromiumos/codesearch/+/main:src/platform/tast-tests/src/chromiumos/tast/local/bundles/cros/video/data/test_vectors/av1/",
        8,
    )
    generator.generate(not args.skip_download, args.jobs)

    generator = ChromiumAV1Generator(
        "chromium-AV1-10bit",
        "CHROMIUM-10bit-AV1-TEST-VECTORS",
        Codec.AV1,
        "AV1 Test Vector Catalogue from https://source.chromium.org/chromiumos/chromiumos/codesearch/+/main:src/platform/tast-tests/src/chromiumos/tast/local/bundles/cros/video/data/test_vectors/av1/",
        10,
    )
    generator.generate(not args.skip_download, args.jobs)



# Fluster - testing framework for decoders conformance
# Copyright (C) 2020, Fluendo, S.A.
#  Author: Pablo Marcos Oltra <pmarcos@fluendo.com>, Fluendo, S.A.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <https://www.gnu.org/licenses/>.

import os
from subprocess import TimeoutExpired
import unittest
from typing import Any
from time import perf_counter

from fluster.decoder import Decoder
from fluster.test_vector import TestVector, TestVectorResult
from fluster.utils import normalize_path


class Test(unittest.TestCase):
    """Test suite for decoder tests"""

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        decoder: Decoder,
        test_suite: Any,  # can't use TestSuite type because of circular dependency
        test_vector: TestVector,
        skip: bool,
        output_dir: str,
        reference: bool,
        timeout: int,
        keep_files: bool,
        verbose: bool,
    ):
        self.decoder = decoder
        self.test_suite = test_suite
        self.test_vector = test_vector
        self.skip = skip
        self.resources_dir = self.test_suite.resources_dir
        self.output_dir = output_dir
        self.reference = reference
        self.timeout = timeout
        self.keep_files = keep_files
        self.verbose = verbose
        setattr(self, test_vector.name, self._test)
        super().__init__(test_vector.name)

    def _test(self) -> None:
        if self.skip:
            self.test_suite.test_vectors[
                self.test_vector.name
            ].test_result = TestVectorResult.NOT_RUN

            return

        output_filepath = os.path.join(self.output_dir, self.test_vector.name + ".out")

        input_filepath = os.path.join(
            self.resources_dir,
            self.test_suite.name,
            self.test_vector.name,
            self.test_vector.input_file,
        )
        output_filepath = normalize_path(output_filepath)
        input_filepath = normalize_path(input_filepath)

        try:
            start = perf_counter()
            result = self.decoder.decode(
                input_filepath,
                output_filepath,
                self.test_vector.output_format,
                self.timeout,
                self.verbose,
                self.keep_files,
            )
            self.test_suite.test_vectors[self.test_vector.name].test_time = (
                perf_counter() - start
            )
        except TimeoutExpired:
            self.test_suite.test_vectors[
                self.test_vector.name
            ].test_result = TestVectorResult.TIMEOUT
            self.test_suite.test_vectors[self.test_vector.name].test_time = (
                perf_counter() - start
            )
            raise
        except Exception:
            self.test_suite.test_vectors[
                self.test_vector.name
            ].test_result = TestVectorResult.ERROR
            self.test_suite.test_vectors[self.test_vector.name].test_time = (
                perf_counter() - start
            )
            raise

        if (
            not self.keep_files
            and os.path.exists(output_filepath)
            and os.path.isfile(output_filepath)
        ):
            os.remove(output_filepath)

        if not self.reference:
            self.test_suite.test_vectors[
                self.test_vector.name
            ].test_result = TestVectorResult.FAIL
            if self.test_vector.result.lower() == result.lower():
                self.test_suite.test_vectors[
                    self.test_vector.name
                ].test_result = TestVectorResult.SUCCESS
            self.assertEqual(
                self.test_vector.result.lower(),
                result.lower(),
                self.test_vector.name,
            )
        else:
            self.test_suite.test_vectors[
                self.test_vector.name
            ].test_result = TestVectorResult.REFERENCE
            self.test_suite.test_vectors[self.test_vector.name].result = result



# Fluster - testing framework for decoders conformance
# Copyright (C) 2023, The ChromiumOS Authors.
#  Author: Alexandre Courbot <acourbot@chromium.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <https://www.gnu.org/licenses/>.

from fluster.codec import Codec, OutputFormat
from fluster.decoder import Decoder, register_decoder
from fluster.utils import file_checksum, run_command


class CrosCodecsDecoder(Decoder):
    """Generic class for cros-codecs decoder"""

    binary = "ccdec"

    def __init__(self) -> None:
        super().__init__()
        self.name = f"ccdec-{self.codec.value}"
        self.description = f"{self.codec.value} cros-codecs decoder"

    def decode(
        self,
        input_filepath: str,
        output_filepath: str,
        output_format: OutputFormat,
        timeout: int,
        verbose: bool,
        keep_files: bool,
    ) -> str:
        """Decodes input_filepath in output_filepath"""

        formats = {
            OutputFormat.YUV420P: "i420",
            OutputFormat.YUV422P: "i422",
            OutputFormat.YUV420P10LE: "i010",
            OutputFormat.YUV420P12LE: "i012",
            OutputFormat.YUV422P10LE: "i210",
            OutputFormat.YUV422P12LE: "i212",
            OutputFormat.YUV444P: "i444",
            OutputFormat.YUV444P10LE: "i410",
            OutputFormat.YUV444P12LE: "i412",
        }

        output_fmt = ""
        try:
            output_fmt = formats[output_format]
        except KeyError as exception:
            raise Exception(f"Unsupported output format {output_format}") from exception

        if self.codec == Codec.H264:
            input_fmt = "h264"
        elif self.codec == Codec.H265:
            input_fmt = "h265"
        elif self.codec == Codec.VP8:
            input_fmt = "vp8"
        elif self.codec == Codec.VP9:
            input_fmt = "vp9"
        elif self.codec == Codec.AV1:
            input_fmt = "av1"
        else:
            raise Exception(f"Unsupported input codec {self.codec}")

        run_command(
            [
                self.binary,
                input_filepath,
                "--output",
                output_filepath,
                "--input-format",
                input_fmt,
                "--output-format",
                output_fmt,
            ],
            timeout=timeout,
            verbose=verbose,
        )

        return file_checksum(output_filepath)


@register_decoder
class CrosCodecsH264Decoder(CrosCodecsDecoder):
    """cros-codecs decoder for H.264"""

    codec = Codec.H264
    hw_acceleration = True
    api = "VAAPI"


@register_decoder
class CrosCodecsH265Decoder(CrosCodecsDecoder):
    """cros-codecs decoder for H.265"""

    codec = Codec.H265
    hw_acceleration = True
    api = "VAAPI"


@register_decoder
class CrosCodecsVp8Decoder(CrosCodecsDecoder):
    """cros-codecs decoder for VP8"""

    codec = Codec.VP8
    hw_acceleration = True
    api = "VAAPI"


@register_decoder
class CrosCodecsVP9Decoder(CrosCodecsDecoder):
    """cros-codecs decoder for VP9"""

    codec = Codec.VP9
    hw_acceleration = True
    api = "VAAPI"


@register_decoder
class CrosCodecsAV1Decoder(CrosCodecsDecoder):
    """cros-codecs decoder for AV1"""

    codec = Codec.AV1
    hw_acceleration = True
    api = "VAAPI"



# Fluster - testing framework for decoders conformance
# Copyright (C) 2020, Fluendo, S.A.
#  Author: Pablo Marcos Oltra <pmarcos@fluendo.com>, Fluendo, S.A.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <https://www.gnu.org/licenses/>.

from enum import Enum
from typing import List, Dict, Type, Any
from fluster.codec import OutputFormat


class TestVectorResult(Enum):
    """Test Result"""

    NOT_RUN = "Not Run"
    SUCCESS = "Success"
    FAIL = "Fail"
    TIMEOUT = "Timeout"
    ERROR = "Error"

    """
    This is only used in reference runs to indicate that
    the decoder for this test vector run succesfully
    """
    REFERENCE = "Reference run"


class TestVector:
    """Test vector"""

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        name: str,
        source: str,
        source_checksum: str,
        input_file: str,
        output_format: OutputFormat,
        result: str,
    ):
        # JSON members
        self.name = name
        self.source = source
        self.source_checksum = source_checksum
        self.input_file = input_file
        self.output_format = output_format
        self.result = result

        # Not included in JSON
        self.test_result = TestVectorResult.NOT_RUN
        self.test_time = 0.0
        self.errors: List[List[str]] = []

    @classmethod
    def from_json(cls: Type["TestVector"], data: Any) -> Any:
        """Deserialize an instance of TestVector from a json file"""
        if "output_format" in data:
            data["output_format"] = OutputFormat(data["output_format"])
        else:
            data["output_format"] = OutputFormat.NONE
        return (data["name"], cls(**data))

    def data_to_serialize(self) -> Dict[str, object]:
        """Return the data to be serialized"""
        data = self.__dict__.copy()
        data.pop("test_result")
        data.pop("errors")
        data.pop("test_time")
        data["output_format"] = str(self.output_format.value)
        return data

    def __str__(self) -> str:
        ret = (
            f"        {self.name}\n"
            f"            Source: {self.source}\n"
            f"            Input: {self.input_file}\n"
            f"            Result: {self.result}"
        )
        return ret



# Fluster - testing framework for decoders conformance
# Copyright (C) 2020, Fluendo, S.A.
#  Author: Pablo Marcos Oltra <pmarcos@fluendo.com>, Fluendo, S.A.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <https://www.gnu.org/licenses/>.

from enum import Enum


class Codec(Enum):
    """Codec type"""

    NONE = "None"
    DUMMY = "Dummy"
    H264 = "H.264"
    H265 = "H.265"
    H266 = "H.266"
    VP8 = "VP8"
    VP9 = "VP9"
    AAC = "AAC"
    AV1 = "AV1"


class OutputFormat(Enum):
    """Output format"""

    NONE = "None"
    YUV420P = "yuv420p"
    YUV422P = "yuv422p"
    YUV420P10LE = "yuv420p10le"
    YUV420P12LE = "yuv420p12le"
    YUV422P10LE = "yuv422p10le"
    YUV422P12LE = "yuv422p12le"
    YUV444P = "yuv444p"
    YUV444P10LE = "yuv444p10le"
    YUV444P12LE = "yuv444p12le"
    YUV444P16LE = "yuv444p16le"
    GRAY = "gray"
    GRAY12LE = "gray12le"
    GRAY16LE = "gray16le"



#!/usr/bin/env python3

# Fluster - testing framework for decoders conformance
# Copyright (C) 2020, Fluendo, S.A.
#  Author: Pablo Marcos Oltra <pmarcos@fluendo.com>, Fluendo, S.A.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <https://www.gnu.org/licenses/>.

import argparse
import os
import multiprocessing
import sys
from importlib import util
from typing import Any, Tuple
from tempfile import gettempdir

from fluster.fluster import Fluster, Context, SummaryFormat
from fluster import utils

APPNAME = "fluster"
TEST_SUITES_DIR = "test_suites"
TEST_SUITES_DIR_SYS = "/usr/share/fluster/test_suites"
DECODERS_DIR = "decoders"
RESOURCES_DIR = "resources"
OUTPUT_DIR = "fluster_output"


def fluster_main() -> None:
    """Entrypoint for the application."""
    main = Main()
    main.run()


class Main:
    """Main class for Fluster"""

    # pylint: disable=broad-except

    def __init__(self) -> None:
        self.decoders_dir = DECODERS_DIR
        self.test_suites_dir = os.path.join(
            os.path.dirname(__file__), "..", TEST_SUITES_DIR
        )
        self.resources_dir = os.path.join(
            os.path.dirname(__file__), "..", RESOURCES_DIR
        )
        self.output_dir = os.path.join(gettempdir(), OUTPUT_DIR)

        is_installed = not os.path.exists(
            os.path.join(os.path.dirname(__file__), "..", ".git")
        )
        if is_installed:
            self.resources_dir, self.test_suites_dir = self._get_installed_dirs()

        self.parser = self._create_parser()

        # Prepend to the PATH the decoders_dir so that we can run them
        # without having to set the env for every single command
        os.environ["PATH"] = self.decoders_dir + os.path.pathsep + os.environ["PATH"]

    def run(self) -> None:
        """Runs Fluster"""
        args = self.parser.parse_args()
        self._validate_args(args)
        self._validate_deps(args)
        if hasattr(args, "func"):
            fluster = Fluster(
                test_suites_dir=args.test_suites_dir,
                decoders_dir=self.decoders_dir,
                resources_dir=args.resources,
                output_dir=args.output,
                use_emoji=not args.no_emoji,
                verbose=args.verbose if "verbose" in args else False,
            )
            args.func(args, fluster)
        else:
            self.parser.print_help()

    def _validate_args(self, args: Any) -> None:
        if hasattr(args, "format"):
            if (
                args.format in [SummaryFormat.JUNITXML.value, SummaryFormat.CSV.value]
                and not args.summary_output
            ):
                sys.exit(
                    "error: please specify XML/CSV file path with -so/--summary-output option."
                )

    def _validate_deps(self, args: Any) -> None:
        if hasattr(args, "format"):
            junit_spec = util.find_spec("junitparser")
            if args.format == SummaryFormat.JUNITXML.value and junit_spec is None:
                sys.exit(
                    "error: junitparser required to use JUnit format. Please install with pip install junitparser."
                )

    def _get_installed_dirs(self) -> Tuple[str, str]:
        site_data_dirs = utils.site_data_dirs(APPNAME)
        user_data_dir = utils.user_data_dir(APPNAME)

        for site_data_dir in site_data_dirs:
            test_suites_dir = os.path.join(site_data_dir, TEST_SUITES_DIR)
            if os.path.exists(test_suites_dir):
                break
        else:
            test_suites_dir = os.path.join(user_data_dir, TEST_SUITES_DIR)

        resources_dir = os.path.join(user_data_dir, RESOURCES_DIR)

        return (resources_dir, test_suites_dir)

    def _create_parser(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-r",
            "--resources",
            help="set the directory where resources are taken from",
            default=self.resources_dir,
        )
        parser.add_argument(
            "-o",
            "--output",
            help="set the directory where decoder outputs will be stored",
            default=self.output_dir,
        )
        parser.add_argument(
            "-ne",
            "--no-emoji",
            help="set to use plain text instead of emojis",
            action="store_true",
        )
        parser.add_argument(
            "-tsd",
            "--test-suites-dir",
            help=(
                "set directory where test suite will be read from, "
                f"multiple directories are supported with OS path separator ({os.pathsep})"
            ),
            default=self.test_suites_dir,
        )
        subparsers = parser.add_subparsers(title="subcommands")
        self._add_list_cmd(subparsers)
        self._add_run_cmd(subparsers)
        self._add_download_cmd(subparsers)
        self._add_reference_cmd(subparsers)
        return parser

    def _add_list_cmd(self, subparsers: Any) -> None:
        subparser = subparsers.add_parser(
            "list",
            aliases=["l"],
            help="show list of available test suites and decoders",
        )
        subparser.add_argument(
            "-ts",
            "--testsuites",
            help="show only the test suites given",
            nargs="+",
        )
        subparser.add_argument(
            "-tv",
            "--testvectors",
            help="show test vectors of test suites",
            action="store_true",
        )
        subparser.add_argument(
            "-c",
            "--check",
            help="check which decoders can be run successfully. Reports ✔️ or ❌",
            action="store_true",
        )
        subparser.add_argument(
            "-v",
            "--verbose",
            help="show stdout and stderr of commands executed",
            action="store_true",
        )
        subparser.set_defaults(func=self._list_cmd)

    def _add_run_cmd(self, subparsers: Any) -> None:
        subparser = subparsers.add_parser(
            "run", aliases=["r"], help="run test suites for decoders"
        )
        subparser.add_argument(
            "-j",
            "--jobs",
            help="number of parallel jobs to use. 1x logical cores by default."
            "0 means all logical cores",
            type=int,
            default=multiprocessing.cpu_count(),
        )
        subparser.add_argument(
            "-t",
            "--timeout",
            help="timeout in secs for each decoding. Defaults to 30 secs",
            type=int,
            default=30,
        )
        subparser.add_argument(
            "-ff",
            "--failfast",
            help="stop after first fail",
            action="store_true",
        )
        subparser.add_argument(
            "-q",
            "--quiet",
            help="don't show every test run",
            action="store_true",
        )
        subparser.add_argument(
            "-ts",
            "--testsuites",
            help="run only the specific test suites",
            nargs="+",
        )
        subparser.add_argument(
            "-tv",
            "--testvectors",
            help="run only the specific test vectors",
            nargs="+",
        )
        subparser.add_argument(
            "-sv",
            "--skipvectors",
            help="skip the specific test vectors",
            nargs="+",
        )
        subparser.add_argument(
            "-d",
            "--decoders",
            help="run only the specific decoders",
            nargs="+",
        )
        subparser.add_argument(
            "-s",
            "--summary",
            help="generate a summary in Markdown format for each test suite",
            action="store_true",
        )
        subparser.add_argument(
            "-so", "--summary-output", help="dump summary output to file"
        )
        subparser.add_argument(
            "-f",
            "--format",
            help="specify the format for the summary file",
            choices=list(x.value for x in SummaryFormat),
            default=SummaryFormat.MARKDOWN.value,
        )
        subparser.add_argument(
            "-k",
            "--keep",
            help="keep output files generated during the test",
            action="store_true",
        )
        subparser.add_argument(
            "-th",
            "--threshold",
            help="set exit code to 2 if threshold tests are not success. "
            "exit code is 0 otherwise",
            type=int,
        )
        subparser.add_argument(
            "-tth",
            "--time-threshold",
            help="set exit code to 3 if test suite takes longer than threshold seconds. "
            "exit code is 0 otherwise",
            type=float,
        )
        subparser.add_argument(
            "-v",
            "--verbose",
            help="show stdout and stderr of commands executed",
            action="store_true",
        )
        subparser.set_defaults(func=self._run_cmd)

    def _add_reference_cmd(self, subparsers: Any) -> None:
        subparser = subparsers.add_parser(
            "reference",
            aliases=["f"],
            help="use a specific decoder to set its results for the test suites given",
        )
        subparser.add_argument(
            "-j",
            "--jobs",
            help="number of parallel jobs to use. 1x logical cores by default."
            "0 means all logical cores",
            type=int,
            default=multiprocessing.cpu_count(),
        )
        subparser.add_argument(
            "-t",
            "--timeout",
            help="timeout in secs for each decoding. Defaults to 30 secs",
            type=int,
            default=30,
        )
        subparser.add_argument("decoder", help="decoder to run", nargs=1)
        subparser.add_argument(
            "testsuites",
            help="list of testsuites to run the decoder with",
            nargs="+",
        )
        subparser.add_argument(
            "-q",
            "--quiet",
            help="don't show every test run",
            action="store_true",
        )
        subparser.add_argument(
            "-v",
            "--verbose",
            help="show stdout and stderr of commands executed",
            action="store_true",
        )
        subparser.set_defaults(func=self._reference_cmd)

    def _add_download_cmd(self, subparsers: Any) -> None:
        subparser = subparsers.add_parser(
            "download", aliases=["d"], help="downloads test suites resources"
        )
        subparser.add_argument(
            "-j",
            "--jobs",
            help="number of parallel jobs to use. 2x logical cores by default."
            "0 means all logical cores",
            type=int,
            default=2 * multiprocessing.cpu_count(),
        )
        subparser.add_argument(
            "-k",
            "--keep",
            help="keep original downloaded file after extracting. Only applicable to compressed "
            "files such as .zip, .tar.gz, etc",
            action="store_true",
        )
        subparser.add_argument(
            "-r",
            "--retries",
            help="number of retries, before failing",
            type=int,
            default=1,
        )
        subparser.add_argument(
            "testsuites", help="list of testsuites to download", nargs="*"
        )
        subparser.set_defaults(func=self._download_cmd)

    def _list_cmd(self, args: Any, fluster: Fluster) -> None:
        fluster.list_test_suites(
            show_test_vectors=args.testvectors, test_suites=args.testsuites
        )
        fluster.list_decoders(check=args.check, verbose=args.verbose)

    def _run_cmd(self, args: Any, fluster: Fluster) -> None:
        args.jobs = args.jobs if args.jobs > 0 else multiprocessing.cpu_count()
        context = Context(
            jobs=args.jobs,
            test_suites=args.testsuites,
            timeout=args.timeout,
            decoders=args.decoders,
            test_vectors=args.testvectors,
            skip_vectors=args.skipvectors,
            failfast=args.failfast,
            quiet=args.quiet,
            summary=args.summary or args.summary_output,
            keep_files=args.keep,
            threshold=args.threshold,
            time_threshold=args.time_threshold,
            verbose=args.verbose,
            summary_output=args.summary_output,
            summary_format=args.format,
        )
        try:
            fluster.run_test_suites(context)
        except SystemExit as exception:
            sys.exit(exception.code)

    def _reference_cmd(self, args: Any, fluster: Fluster) -> None:
        context = Context(
            jobs=args.jobs,
            timeout=args.timeout,
            test_suites=args.testsuites,
            decoders=args.decoder,
            test_vectors=[],
            skip_vectors=[],
            quiet=args.quiet,
            verbose=args.verbose,
            reference=True,
        )
        try:
            fluster.run_test_suites(context)
        except SystemExit as exception:
            sys.exit(exception.code)

    def _download_cmd(self, args: Any, fluster: Fluster) -> None:
        args.jobs = args.jobs if args.jobs > 0 else multiprocessing.cpu_count()
        fluster.download_test_suites(
            test_suites=args.testsuites,
            jobs=args.jobs,
            keep_file=args.keep,
            retries=args.retries,
        )



# Fluster - testing framework for decoders conformance
# Copyright (C) 2020, Fluendo, S.A.
#  Author: Pablo Marcos Oltra <pmarcos@fluendo.com>, Fluendo, S.A.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <https://www.gnu.org/licenses/>.

import glob
import os.path

modules = glob.glob(os.path.join(os.path.dirname(__file__), "*.py"))
__all__ = [
    os.path.basename(os.path.splitext(f)[0])
    for f in modules
    if os.path.isfile(f) and not f.endswith("__init__.py")
]



#!/usr/bin/env python3

# Fluster - testing framework for decoders conformance
# Copyright (C) 2020, Fluendo, S.A.
#  Author: Pablo Marcos Oltra <pmarcos@fluendo.com>, Fluendo, S.A.
#  Author: Andoni Morales Alastruey <amorales@fluendo.com>, Fluendo, S.A.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <https://www.gnu.org/licenses/>.

import argparse
from html.parser import HTMLParser
import os
import sys
import urllib.request
import multiprocessing
import re

# pylint: disable=wrong-import-position
sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
from fluster import utils
from fluster.codec import Codec, OutputFormat
from fluster.test_suite import TestSuite, TestVector

# pylint: enable=wrong-import-position

BASE_URL = "https://www.itu.int/"
H266_URL = BASE_URL + "wftp3/av-arch/jvet-site/bitstream_exchange/VVC/draft_conformance/"
H265_URL = BASE_URL + "wftp3/av-arch/jctvc-site/bitstream_exchange/draft_conformance/"
H264_URL = BASE_URL + "wftp3/av-arch/jvt-site/draft_conformance/"
BITSTREAM_EXTS = (
    ".bin",
    ".bit",
    ".264",
    ".h264",
    ".jvc",
    ".jsv",
    ".jvt",
    ".avc",
    ".26l",
)
MD5_EXTS = ("yuv_2.md5", "yuv.md5", ".md5", "md5.txt", "md5sum.txt")
MD5_EXCLUDES = (".bin.md5", "bit.md5")
RAW_EXTS = ("nogray.yuv", ".yuv", ".qcif")


class HREFParser(HTMLParser):
    """Custom parser to find href links"""

    def __init__(self):
        self.links = []
        super().__init__()

    def error(self, message):
        print(message)

    def handle_starttag(self, tag, attrs):
        # Only parse the 'anchor' tag.
        if tag == "a":
            # Check the list of defined attributes.
            for name, value in attrs:
                # If href is defined, print it.
                if name == "href":
                    base_url = BASE_URL if BASE_URL[-1] != "/" else BASE_URL[0:-1]
                    self.links.append(base_url + value)


class JCTVTGenerator:
    """Generates a test suite from the conformance bitstreams"""

    def __init__(
        self,
        name: str,
        suite_name: str,
        codec: Codec,
        description: str,
        site: str,
        use_ffprobe: bool = False
    ):
        self.name = name
        self.suite_name = suite_name
        self.codec = codec
        self.description = description
        self.site = site
        self.use_ffprobe = use_ffprobe

    def generate(self, download, jobs):
        """Generates the test suite and saves it to a file"""
        output_filepath = os.path.join(self.suite_name + ".json")
        test_suite = TestSuite(
            output_filepath,
            "resources",
            self.suite_name,
            self.codec,
            self.description,
            dict(),
        )

        hparser = HREFParser()
        print(f"Download list of bitstreams from {self.site + self.name}")
        with urllib.request.urlopen(self.site + self.name) as resp:
            data = str(resp.read())
            hparser.feed(data)

        for url in hparser.links[1:]:
            # The first item in the AVCv1 list is a readme file
            if "00readme_H" in url:
                continue
            elif "replaced" in url:
                # This is in HEVC-SHVC, we don't want that.
                continue
            file_url = os.path.basename(url)
            name = os.path.splitext(file_url)[0]
            file_input = f"{name}.bin"
            test_vector = TestVector(name, url, "__skip__", file_input, OutputFormat.YUV420P, "")
            test_suite.test_vectors[name] = test_vector

        if download:
            test_suite.download(
                jobs=jobs,
                out_dir=test_suite.resources_dir,
                verify=False,
                extract_all=True,
                keep_file=True,
            )

        for test_vector in test_suite.test_vectors.values():
            dest_dir = os.path.join(
                test_suite.resources_dir, test_suite.name, test_vector.name
            )
            dest_path = os.path.join(dest_dir, os.path.basename(test_vector.source))
            test_vector.input_file = self._find_by_ext(dest_dir, BITSTREAM_EXTS)
            absolute_input_path = test_vector.input_file
            test_vector.input_file = test_vector.input_file.replace(
                os.path.join(
                    test_suite.resources_dir, test_suite.name, test_vector.name
                )
                + os.sep,
                "",
            )
            if not test_vector.input_file:
                raise Exception(f"Bitstream file not found in {dest_dir}")
            test_vector.source_checksum = utils.file_checksum(dest_path)
            if "main10" in test_vector.name.lower():
                test_vector.output_format = OutputFormat.YUV420P10LE
            elif self.use_ffprobe:
                ffprobe = utils.normalize_binary_cmd('ffprobe')
                command = [ffprobe, '-v', 'error', '-select_streams', 'v:0',
                           '-show_entries', 'stream=pix_fmt', '-of',
                           'default=nokey=1:noprint_wrappers=1',
                           absolute_input_path]

                result = utils.run_pipe_command_with_std_output(command)
                pix_fmt = result[0]
                try:
                    test_vector.output_format = OutputFormat[pix_fmt.upper()]
                except KeyError as e:
                    exceptions = {
                        # Feature: Test unequal luma and chroma bitdepth
                        # setting. The luma bitdepth is higher than the chroma
                        # bitdepth. Luma is 12bit, chroma is 8bit. Considering
                        # 12bit.
                        "Bitdepth_A_RExt_Sony_1": OutputFormat.YUV444P12LE,
                        # Same as above, but the chroma is 12bit and luma is 8bit.
                        "Bitdepth_B_RExt_Sony_1": OutputFormat.YUV444P12LE,
                        # Rest is taken by examining the error displayed by ffprobe, e.g.:
                        # The following bit-depths are currently specified: 8,
                        # 9, 10 and 12 bits, chroma_format_idc is 3, depth is 16
                        "EXTPREC_MAIN_444_16_INTRA_10BIT_RExt_Sony_1": OutputFormat.YUV444P16LE,
                        "EXTPREC_HIGHTHROUGHPUT_444_16_INTRA_16BIT_RExt_Sony_1": OutputFormat.YUV444P16LE,
                        "EXTPREC_MAIN_444_16_INTRA_16BIT_RExt_Sony_1": OutputFormat.YUV444P16LE,
                        "GENERAL_16b_400_RExt_Sony_1" : OutputFormat.GRAY16LE,
                        "GENERAL_16b_444_highThroughput_RExt_Sony_2" : OutputFormat.YUV444P16LE,
                        "GENERAL_16b_444_RExt_Sony_2": OutputFormat.YUV444P16LE,
                        "WAVETILES_RExt_Sony_2" : OutputFormat.YUV444P16LE
                    }
                    if test_vector.name in exceptions.keys():
                        test_vector.output_format = exceptions[test_vector.name]
                    else:
                        raise e


            if self.codec == Codec.H265:
                self._fill_checksum_h265(test_vector, dest_dir)
            elif self.codec == Codec.H264:
                self._fill_checksum_h264(test_vector, dest_dir)

        test_suite.to_json_file(output_filepath)
        print("Generate new test suite: " + test_suite.name + ".json")

    def _fill_checksum_h264(self, test_vector, dest_dir):
        raw_file = self._find_by_ext(dest_dir, RAW_EXTS)
        if raw_file is None:
            raise Exception(f"RAW file not found in {dest_dir}")
        test_vector.result = utils.file_checksum(raw_file)

    def _fill_checksum_h265(self, test_vector, dest_dir):
        checksum_file = self._find_by_ext(dest_dir, MD5_EXTS, MD5_EXCLUDES)
        if checksum_file is None:
            raise Exception("MD5 not found")
        with open(checksum_file, "r") as checksum_file:
            # The md5 is in several formats
            # Example 1
            # 158312a1a35ef4b20cb4aeee48549c03 *WP_A_Toshiba_3.bit
            # Example 2
            # MD5 (rec.yuv) = e5c4c20a8871aa446a344efb1755bcf9
            # Example 3
            # # MD5 checksums generated by MD5summer (http://www.md5summer.org)
            # # Generated 6/14/2013 4:22:11 PM
            # 29799285628de148502da666a7fc2df5 *DBLK_F_VIXS_1.bit
            # Example 4:
            # cd84e70999e4577f23eb2077c7a1d24f  EXTPREC_HIGHTHROUGHPUT_444_16_INTRA_8BIT_RExt_Sony_1_rec.yuv
            # 812d62fa8566d51544e760fd9ecbe138  EXTPREC_HIGHTHROUGHPUT_444_16_INTRA_8BIT_RExt_Sony_1.bit
            # Example 5 (extra 4 bits at the end, possible typo?)
            # 580e4041563c9cca6c30d0b6c09571aef *Slice_ACT_QP_Offsets_A_Qualcomm_2.bit
            # 2d4ffc0354dee7216567c89dc1f20391 *Slice_ACT_QP_Offsets_A_Qualcomm_2.rgb
            # Example 6:
            # 9cab6bcd74491062a8523b5a7ff6a540  CCP_8bit_RExt_QCOM.bin
            # f3e914fccdb820eac85f46642ea0e168  CCP_8bit_RExt_QCOM.gbr
            regex = re.compile(rf"([a-fA-F0-9]{{32,}}).*\.(yuv|rgb|gbr)")
            lines = checksum_file.readlines()
            # If we have a line like examples 4,5,6 anywhere in the file, prefer
            # that.
            if self.codec == Codec.H265 and any((match := regex.match(line)) for line in lines):
                test_vector.result = match.group(1)[:32].lower()
            elif self.codec == Codec.H265 and self.name == "RExt" or self.name == "MV-HEVC" or self.name == "SCC":
                # If we can't match with the regex, note that these usually come
                # with the checksum at the end
                test_vector.result = lines[-1].split(" ")[0].split("\n")[0].lower()
            else:
                for line in lines:
                    if line.startswith(("#", "\n")):
                        continue
                    if "=" in line:
                        test_vector.result = line.split("=")[-1].strip().lower()
                    else:
                        test_vector.result = line.split(" ")[0].split("\n")[0].lower()
                    break
            # Assert that we have extracted a valid MD5 from the file
            assert len(test_vector.result) == 32 and re.search(
                r"^[a-fA-F0-9]{32}$", test_vector.result) != None, f"{test_vector.result} is not a valid MD5 hash"

    def _find_by_ext(self, dest_dir, exts, excludes=None):
        excludes = excludes or []

        # Respect the priority for extensions
        for ext in exts:
            for subdir, _, files in os.walk(dest_dir):
                for filename in files:
                    excluded = False
                    filepath = subdir + os.sep + filename
                    if not filepath.endswith(ext) or "__MACOSX" in filepath:
                        continue
                    for excl in excludes:
                        if excl in filepath:
                            excluded = True
                            break
                    if not excluded:
                        return filepath
        return None


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--skip-download",
        help="skip extracting tarball",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-j",
        "--jobs",
        help="number of parallel jobs to use. 2x logical cores by default",
        type=int,
        default=2 * multiprocessing.cpu_count(),
    )
    args = parser.parse_args()
    generator = JCTVTGenerator(
        "HEVC_v1",
        "JCT-VC-HEVC_V1",
        Codec.H265,
        "JCT-VC HEVC version 1",
        H265_URL,
    )
    generator.generate(not args.skip_download, args.jobs)

    generator = JCTVTGenerator(
        "RExt",
        "JCT-VC-RExt",
        Codec.H265,
        "JCT-VC HEVC Range Extension",
        H265_URL,
        True
    )
    generator.generate(not args.skip_download, args.jobs)

    generator = JCTVTGenerator(
        "SCC",
        "JCT-VC-SCC",
        Codec.H265,
        "JCT-VC HEVC Screen Content Coding Extension",
        H265_URL,
        True
    )
    generator.generate(not args.skip_download, args.jobs)

    generator = JCTVTGenerator(
        "MV-HEVC",
        "JCT-VC-MV-HEVC",
        Codec.H265,
        "JCT-VC HEVC Multiview Extension",
        H265_URL,
        True
    )
    generator.generate(not args.skip_download, args.jobs)

    generator = JCTVTGenerator(
        "AVCv1", "JVT-AVC_V1", Codec.H264, "JVT AVC version 1", H264_URL
    )
    generator.generate(not args.skip_download, args.jobs)

    generator = JCTVTGenerator('draft6', 'JVET-VVC_draft6', Codec.H266,
                               'JVET VVC draft6', H266_URL)
    generator.generate(not args.skip_download, args.jobs)



# Fluster - testing framework for decoders conformance
# Copyright (C) 2020, Fluendo, S.A.
#  Author: Pablo Marcos Oltra <pmarcos@fluendo.com>, Fluendo, S.A.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <https://www.gnu.org/licenses/>.

from collections import defaultdict
import os
import os.path
from functools import lru_cache
from typing import List, Dict, Any, Tuple, Optional, Iterator
import sys
from enum import Enum
from shutil import rmtree

# Import decoders that will auto-register
# pylint: disable=wildcard-import, unused-wildcard-import
from fluster.decoders import *  # noqa: F401,F403

# pylint: enable=wildcard-import, unused-wildcard-import

from fluster.test_suite import TestSuite
from fluster.test_suite import Context as TestSuiteContext
from fluster.decoder import DECODERS, Decoder
from fluster.test_vector import TestVector, TestVectorResult
from fluster.codec import Codec

# pylint: disable=broad-except


class Context:
    """Context for run and reference command"""

    # pylint: disable=too-many-instance-attributes too-many-locals

    def __init__(
        self,
        jobs: int,
        timeout: int,
        test_suites: List[str],
        decoders: List[str],
        test_vectors: List[str],
        skip_vectors: List[str],
        failfast: bool = False,
        quiet: bool = False,
        reference: bool = False,
        summary: bool = False,
        keep_files: bool = False,
        threshold: Optional[int] = None,
        time_threshold: Optional[int] = None,
        verbose: bool = False,
        summary_output: str = "",
        summary_format: str = "",
    ):
        self.jobs = jobs
        self.timeout = timeout
        self.test_suites_names = test_suites
        self.test_suites: List[TestSuite] = []
        self.decoders_names = decoders
        self.decoders: List[Decoder] = []
        self.test_vectors_names = test_vectors
        self.skip_vectors_names = skip_vectors
        self.failfast = failfast
        self.quiet = quiet
        self.reference = reference
        self.summary = summary
        self.keep_files = keep_files
        self.threshold = threshold
        self.time_threshold = time_threshold
        self.verbose = verbose
        self.summary_output = summary_output
        self.summary_format = summary_format

    def to_test_suite_context(
        self,
        decoder: Decoder,
        output_dir: str,
        test_vectors: List[str],
        skip_vectors: List[str],
    ) -> TestSuiteContext:
        """Create a TestSuite's Context from this"""
        ts_context = TestSuiteContext(
            jobs=self.jobs,
            decoder=decoder,
            timeout=self.timeout,
            failfast=self.failfast,
            quiet=self.quiet,
            output_dir=output_dir,
            reference=self.reference,
            test_vectors=test_vectors,
            skip_vectors=skip_vectors,
            keep_files=self.keep_files,
            verbose=self.verbose,
        )
        return ts_context


EMOJI_RESULT = {
    TestVectorResult.NOT_RUN: "",
    TestVectorResult.SUCCESS: "✔️",
    TestVectorResult.FAIL: "❌",
    TestVectorResult.TIMEOUT: "⌛",
    TestVectorResult.ERROR: "☠",
}

TEXT_RESULT = {
    TestVectorResult.NOT_RUN: "",
    TestVectorResult.SUCCESS: "OK",
    TestVectorResult.FAIL: "KO",
    TestVectorResult.TIMEOUT: "TO",
    TestVectorResult.ERROR: "ER",
}


class SummaryFormat(Enum):
    """Summary formats"""

    MARKDOWN = "md"
    CSV = "csv"
    JUNITXML = "junitxml"


class Fluster:
    """Main class for fluster"""

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        test_suites_dir: str,
        decoders_dir: str,
        resources_dir: str,
        output_dir: str,
        verbose: bool = False,
        use_emoji: bool = True,
    ):
        self.test_suites_dir = test_suites_dir
        self.decoders_dir = decoders_dir
        self.resources_dir = resources_dir
        self.output_dir = output_dir
        self.verbose = verbose
        self.test_suites: List[TestSuite] = []
        self.decoders = DECODERS
        self.emoji = EMOJI_RESULT if use_emoji else TEXT_RESULT
        if self.verbose:
            print(
                f"NOTE: Internal dirs used:\n"
                f" * test_suites_dir: {self.test_suites_dir}\n"
                f" * resources_dir: {self.resources_dir}\n"
                f" * output_dir: {self.output_dir}"
            )

    def _walk_test_suite_dir(self) -> Iterator[Tuple[str, List[str], List[str]]]:
        for test_suite_dir in self.test_suites_dir.split(os.pathsep):
            for root, dirnames, files in os.walk(test_suite_dir):
                yield (root, dirnames, files)

    @lru_cache(maxsize=128)
    def _load_test_suites(self) -> None:
        for root, _, files in self._walk_test_suite_dir():
            for file in files:
                if os.path.splitext(file)[1] == ".json":
                    try:
                        test_suite = TestSuite.from_json_file(
                            os.path.join(root, file), self.resources_dir
                        )
                        if test_suite.name in [ts.name for ts in self.test_suites]:
                            raise Exception(
                                f'Repeated test suite with name "{test_suite.name}"'
                            )
                        self.test_suites.append(test_suite)
                    except Exception as ex:
                        print(f"Error loading test suite {file}: {ex}")
        if len(self.test_suites) == 0:
            raise Exception(f'No test suites found in "{self.test_suites_dir}"')

    def list_decoders(self, check: bool, verbose: bool) -> None:
        """List all the available decoders"""
        print("\nList of available decoders:")
        decoders_dict: Dict[Codec, List[Decoder]] = {}
        for dec in self.decoders:
            if dec.codec not in decoders_dict:
                decoders_dict[dec.codec] = []
            decoders_dict[dec.codec].append(dec)

        for codec, decoder_list in decoders_dict.items():
            print(f'\n{str(codec).split(".")[1]}')
            for decoder in decoder_list:
                string = f"{decoder}"
                if check:
                    string += "... " + (
                        self.emoji[TestVectorResult.SUCCESS]
                        if decoder.check(verbose)
                        else self.emoji[TestVectorResult.FAIL]
                    )
                print(string)

    def list_test_suites(
        self,
        show_test_vectors: bool = False,
        test_suites: Optional[List[str]] = None,
    ) -> None:
        """List all test suites"""
        self._load_test_suites()
        print("\nList of available test suites:")
        if test_suites:
            test_suites = [x.lower() for x in test_suites]
        for test_suite in self.test_suites:
            if test_suites:
                if test_suite.name.lower() not in test_suites:
                    continue
            print(test_suite)
            if show_test_vectors:
                for test_vector in test_suite.test_vectors.values():
                    print(test_vector)
        if len(self.test_suites) == 0:
            print(f'    No test suites found in "{self.test_suites_dir}"')

    def _get_matches(
        self, in_list: List[str], check_list: List[Any], name: str
    ) -> List[Any]:
        if in_list:
            in_list_names = {x.lower() for x in in_list}
            check_list_names = {x.name.lower() for x in check_list}
            matches = in_list_names & check_list_names
            if len(matches) != len(in_list):
                sys.exit(
                    f'No {name} found for: {", ".join(in_list_names - check_list_names)}'
                )
            matches_ret = [x for x in check_list if x.name.lower() in matches]
        else:
            matches_ret = check_list
        return matches_ret

    def _normalize_context(self, ctx: Context) -> None:
        # Convert all test suites and decoders to lowercase to make the filter greedy
        if ctx.test_suites:
            ctx.test_suites_names = [x.lower() for x in ctx.test_suites_names]
        if ctx.decoders_names:
            ctx.decoders_names = [x.lower() for x in ctx.decoders_names]
        if ctx.test_vectors_names:
            ctx.test_vectors_names = [x.lower() for x in ctx.test_vectors_names]
        if ctx.skip_vectors_names:
            ctx.skip_vectors_names = [x.lower() for x in ctx.skip_vectors_names]
        ctx.test_suites = self._get_matches(
            ctx.test_suites_names, self.test_suites, "test suite"
        )
        ctx.decoders = self._get_matches(ctx.decoders_names, self.decoders, "decoders")

    def run_test_suites(self, ctx: Context) -> None:
        """Run a group of test suites"""
        # pylint: disable=too-many-branches

        self._load_test_suites()
        self._normalize_context(ctx)

        if ctx.reference and (not ctx.decoders or len(ctx.decoders) > 1):
            dec_names = [dec.name for dec in ctx.decoders]
            raise Exception(
                f'Only one decoder can be the reference. Given: {", ".join(dec_names)}'
            )

        if ctx.threshold and len(ctx.test_suites) > 1:
            raise Exception(
                "Threshold for success tests can only be applied running a single test "
                "suite for a single decoder"
            )

        if ctx.reference:
            print("\n=== Reference mode ===\n")

        error = False
        no_test_run = True
        results: Dict[str, List[Tuple[Decoder, TestSuite]]] = {}
        for test_suite in ctx.test_suites:
            test_suite_results: List[Tuple[Decoder, TestSuite]] = []
            for decoder in ctx.decoders:
                if decoder.codec != test_suite.codec:
                    continue
                test_suite_res = test_suite.run(
                    ctx.to_test_suite_context(
                        decoder,
                        self.output_dir,
                        ctx.test_vectors_names,
                        ctx.skip_vectors_names,
                    )
                )

                if test_suite_res:
                    no_test_run = False
                    test_suite_results.append((decoder, test_suite_res))
                    results[test_suite.name] = test_suite_results
                    success = True
                    for test_vector in test_suite_res.test_vectors.values():
                        if test_vector.errors:
                            success = False
                            break

                    if not success:
                        error = True
                        if ctx.failfast:
                            self._show_summary_if_needed(ctx, results)
                            sys.exit(1)

                    if ctx.threshold:
                        if test_suite_res.test_vectors_success < ctx.threshold:
                            self._show_summary_if_needed(ctx, results)
                            print(
                                f"Tests results below threshold: {test_suite_res.test_vectors_success} vs "
                                f"{ctx.threshold}\nReporting error through exit code 2"
                            )
                            sys.exit(2)

                    if ctx.time_threshold:
                        if test_suite_res.time_taken > ctx.time_threshold:
                            self._show_summary_if_needed(ctx, results)
                            print(
                                f"Tests results over time threshold: {test_suite_res.time_taken} vs "
                                f"{ctx.time_threshold}\nReporting error through exit code 3"
                            )
                            sys.exit(3)

        self._show_summary_if_needed(ctx, results)

        if not ctx.keep_files and os.path.isdir(self.output_dir):
            rmtree(self.output_dir)

        if (error and (not ctx.threshold and not ctx.time_threshold)) or no_test_run:
            sys.exit(1)

    def _show_summary_if_needed(
        self, ctx: Context, results: Dict[str, List[Tuple[Decoder, TestSuite]]]
    ) -> None:
        if ctx.summary and results:
            if ctx.summary_format == SummaryFormat.JUNITXML.value:
                self._generate_junit_summary(ctx, results)
            elif ctx.summary_format == SummaryFormat.CSV.value:
                self._generate_csv_summary(ctx, results)
            else:
                self._generate_md_summary(ctx, results)

    def _generate_junit_summary(
        self, ctx: Context, results: Dict[str, List[Tuple[Decoder, TestSuite]]]
    ) -> None:
        # pylint: disable=import-outside-toplevel

        try:
            import junitparser as junitp  # type: ignore
        except ImportError:
            sys.exit(
                "error: junitparser required to use JUnit format. Please install with pip install junitparser."
            )

        def _parse_vector_errors(vector: TestVector) -> List[junitp.Error]:
            junit_err_map = {
                TestVectorResult.ERROR: junitp.Error,
                TestVectorResult.FAIL: junitp.Failure,
                TestVectorResult.NOT_RUN: junitp.Skipped,
                TestVectorResult.TIMEOUT: junitp.Failure,
            }

            jerrors = []

            for err in vector.errors:
                jerr = junit_err_map[vector.test_result](message=f"FAIL: {err[0]}")
                jerr.text = "\n".join(err[1:])
                jerrors.append(jerr)

            return jerrors

        def _parse_suite_results(
            test_suite_tuple: Tuple[str, List[Tuple[Decoder, TestSuite]]]
        ) -> junitp.TestSuite:
            jsuites = []

            test_suite_name, test_suite_results = test_suite_tuple

            for suite_decoder_res in test_suite_results:
                timeouts = 0

                jsuite = junitp.TestSuite(test_suite_name)
                jsuite.add_property("decoder", suite_decoder_res[0].name)

                for vector in suite_decoder_res[1].test_vectors.values():
                    jcase = junitp.TestCase(vector.name)
                    if vector.test_result == TestVectorResult.NOT_RUN:
                        jcase.result = [junitp.Skipped()]
                    elif vector.test_result not in [
                        TestVectorResult.SUCCESS,
                        TestVectorResult.REFERENCE,
                    ]:
                        jcase.result = _parse_vector_errors(vector)

                    jcase.time = vector.test_time

                    jsuite.add_testcase(jcase)

                    if vector.test_result is TestVectorResult.TIMEOUT and ctx.jobs == 1:
                        timeouts += ctx.timeout

                jsuite.time = round(suite_decoder_res[1].time_taken - timeouts, 3)

                jsuites.append(jsuite)

            return jsuites

        xml = junitp.JUnitXml()

        jsuites = map(_parse_suite_results, results.items())

        for jsuite in [item for sublist in jsuites for item in sublist]:
            xml.add_testsuite(jsuite)

        if ctx.summary_output:
            with open(ctx.summary_output, "w+", encoding="utf-8") as summary_file:
                xml.write(summary_file.name, pretty=True)

    def _generate_csv_summary(
        self, ctx: Context, results: Dict[str, List[Tuple[Decoder, TestSuite]]]
    ) -> None:
        # pylint: disable=too-many-locals
        result_map = {
            TestVectorResult.SUCCESS: "Success",
            TestVectorResult.REFERENCE: "Reference",
            TestVectorResult.TIMEOUT: "Timeout",
            TestVectorResult.ERROR: "Error",
            TestVectorResult.FAIL: "Fail",
            TestVectorResult.NOT_RUN: "Not run",
        }
        content: Dict[Any, Any] = defaultdict(lambda: defaultdict(dict))
        max_vectors = 0
        for test_suite, suite_results in results.items():
            for decoder, vectors in suite_results:
                decoder_name = str(decoder.name[: decoder.name.find(":")])
                max_vectors = max(max_vectors, len(vectors.test_vectors.values()))
                for vector in vectors.test_vectors.values():
                    vector_name = str(vector.name)
                    content[str(test_suite)][decoder_name][vector_name] = result_map[
                        vector.test_result
                    ]

        suite_row = []
        decoder_row = []
        field_row = []
        content_rows: List[List[Any]] = [[] for _ in range(max_vectors)]
        for suite in content:
            suite_row.append(str(suite))
            num_decoders = len(content[suite])
            suite_row += ["" for _ in range(num_decoders + (num_decoders - 1))]
            for decoder in content[suite]:
                decoder_row += [str(decoder), ""]
                field_row += ["Vector", "Result"]
                for index, vector in enumerate(content[suite][decoder]):
                    content_rows[index] += [vector, content[suite][decoder][vector]]
                for index in range(len(content[suite][decoder]), max_vectors):
                    content_rows[index] += ["", ""]
        rows = [suite_row, decoder_row, field_row] + content_rows
        if ctx.summary_output:
            with open(ctx.summary_output, mode="w", encoding="utf8") as file:
                file.writelines([",".join(row) + "\n" for row in rows])

    def _generate_md_summary(
        self, ctx: Context, results: Dict[str, List[Tuple[Decoder, TestSuite]]]
    ) -> None:
        def _global_stats(
            results: List[Tuple[Decoder, TestSuite]],
            test_suites: List[TestSuite],
            first: bool,
        ) -> str:
            separator = f'\n|-|{"-|" * len(results)}'
            output = separator if not first else ""
            output += "\n|Test|" if not first else "|Test|"
            for decoder, _ in results:
                output += f"{decoder.name}|"
            output += separator if first else ""
            output += "\n|TOTAL|"
            for test_suite in test_suites:
                output += (
                    f"{test_suite.test_vectors_success}/{len(test_suite.test_vectors)}|"
                )
            output += "\n|TOTAL TIME|"
            for test_suite in test_suites:
                # Substract from the total time that took running a test suite on a decoder
                # the timeouts. This is not ideal since we won't be comparing decoding the
                # same number of test vectors, but at least it is much better than comparing
                # total times when timeouts are such a huge part of the global time taken.
                # Note: we only do this when the number of parallel jobs is 1, because
                # whenever there are actual parallel jobs, this gets much more complicated.
                timeouts = (
                    sum(
                        [
                            ctx.timeout
                            for tv in test_suite.test_vectors.values()
                            if tv.test_result == TestVectorResult.TIMEOUT
                        ]
                    )
                    if ctx.jobs == 1
                    else 0
                )
                total_time = test_suite.time_taken - timeouts
                output += f"{total_time:.3f}s|"
            output += separator if first else ""
            return output

        output = ""

        for test_suite_name, test_suite_results in results.items():
            decoders_names = [decoder.name for decoder, _ in test_suite_results]
            test_suites = [res[1] for res in test_suite_results]
            print(
                f'Generating summary for test suite {test_suite_name} and decoders {", ".join(decoders_names)}:\n'
            )
            output += _global_stats(test_suite_results, test_suites, True)
            for test_vector in test_suite_results[0][1].test_vectors.values():
                output += f"\n|{test_vector.name}|"
                for test_suite in test_suites:
                    tvector = test_suite.test_vectors[test_vector.name]
                    output += self.emoji[tvector.test_result] + "|"
            output += _global_stats(test_suite_results, test_suites, False)
            output += "\n\n"
        if ctx.summary_output:
            with open(ctx.summary_output, "w+", encoding="utf-8") as summary_file:
                summary_file.write(output)
        else:
            print(output)

    def download_test_suites(
        self, test_suites: List[str], jobs: int, keep_file: bool, retries: int
    ) -> None:
        """Download a group of test suites"""
        self._load_test_suites()
        if not test_suites:
            download_test_suites = self.test_suites
        else:
            download_test_suites = self._get_matches(
                test_suites, self.test_suites, "test suites"
            )

        for test_suite in download_test_suites:
            test_suite.download(
                jobs,
                self.resources_dir,
                verify=True,
                keep_file=keep_file,
                retries=retries,
            )



# Fluster - testing framework for decoders conformance
# Copyright (C) 2020, Fluendo, S.A.
#  Author: Pablo Marcos Oltra <pmarcos@fluendo.com>, Fluendo, S.A.
#  Author: Andoni Morales Alastruey <amorales@fluendo.com>, Fluendo, S.A.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <https://www.gnu.org/licenses/>.

from fluster.codec import Codec, OutputFormat
from fluster.decoder import Decoder, register_decoder
from fluster.utils import file_checksum, run_command


@register_decoder
class H265JCTVTDecoder(Decoder):
    """JCT-VT H.265/HEVC reference decoder implementation"""

    name = "JCT-VT-H.265"
    description = "JCT-VT H.265/HEVC reference decoder"
    codec = Codec.H265
    binary = "TAppDecoder"

    def decode(
        self,
        input_filepath: str,
        output_filepath: str,
        output_format: OutputFormat,
        timeout: int,
        verbose: bool,
        keep_files: bool,
    ) -> str:
        """Decodes input_filepath in output_filepath"""
        # pylint: disable=unused-argument
        run_command(
            [self.binary, "-b", input_filepath, "-o", output_filepath],
            timeout=timeout,
            verbose=verbose,
        )
        return file_checksum(output_filepath)



# Fluster - testing framework for decoders conformance
# Copyright (C) 2020, Fluendo, S.A.
#  Author: Pablo Marcos Oltra <pmarcos@fluendo.com>, Fluendo, S.A.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from functools import lru_cache
from shutil import which
from typing import List, Type
from fluster.codec import OutputFormat, Codec
from fluster.utils import normalize_binary_cmd

# pylint: disable=broad-except


class Decoder(ABC):
    """Base class for decoders"""

    name = ""
    codec = Codec.NONE
    hw_acceleration = False
    description = ""
    binary = ""

    def __init__(self) -> None:
        if self.binary:
            self.binary = normalize_binary_cmd(self.binary)

    @abstractmethod
    def decode(
        self,
        input_filepath: str,
        output_filepath: str,
        output_format: OutputFormat,
        timeout: int,
        verbose: bool,
        keep_files: bool,
    ) -> str:
        """Decodes input_filepath in output_filepath"""
        raise Exception("Not implemented")

    @lru_cache(maxsize=128)
    def check(self, verbose: bool) -> bool:
        """Checks whether the decoder can be run"""
        if hasattr(self, "binary") and self.binary:
            try:
                path = which(self.binary)
                if verbose and not path:
                    print(f"Binary {self.binary} can't be found to be executed")

                return path is not None
            except Exception:
                return False
        return True

    def __str__(self) -> str:
        return f"    {self.name}: {self.description}"


DECODERS: List[Decoder] = []


def register_decoder(cls: Type[Decoder]) -> Type[Decoder]:
    """Register a new decoder implementation"""
    DECODERS.append(cls())
    DECODERS.sort(key=lambda dec: dec.name)
    return cls



# Fluster - testing framework for decoders conformance
# Copyright (C) 2020, Fluendo, S.A.
#  Author: Pablo Marcos Oltra <pmarcos@fluendo.com>, Fluendo, S.A.
#  Author: Andoni Morales Alastruey <amorales@fluendo.com>, Fluendo, S.A.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <https://www.gnu.org/licenses/>.

import os
from functools import lru_cache
from typing import List, Optional, Match
import shlex
import subprocess
import re

from fluster.codec import Codec, OutputFormat
from fluster.decoder import Decoder, register_decoder
from fluster.utils import file_checksum, run_command

FFMPEG_TPL = "{} -i {} {} -vf {}format=pix_fmts={} -f rawvideo {}"


def output_format_to_ffformat(output_format: OutputFormat) -> str:
    """Return GStreamer pixel format"""
    mapping = {
        OutputFormat.YUV420P: "nv12",
        OutputFormat.YUV422P: "nv12",  # vulkan
        OutputFormat.YUV420P10LE: "p010",
        OutputFormat.YUV422P10LE: "p012",
    }
    if output_format not in mapping:
        raise Exception(
            f"No matching output format found in FFmpeg for {output_format}"
        )
    return mapping[output_format]


class FFmpegDecoder(Decoder):
    """Generic class for FFmpeg decoder"""

    binary = "ffmpeg"
    description = ""
    cmd = ""
    api = ""
    wrapper = False
    hw_download = False
    init_device = ""

    def __init__(self) -> None:
        super().__init__()
        self.cmd = self.binary
        if self.hw_acceleration:
            if self.init_device:
                self.cmd += f' -init_hw_device "{self.init_device}" -hwaccel_output_format {self.api.lower()}'
            if self.wrapper:
                self.cmd += f" -c:v {self.api.lower()}"
            else:
                self.cmd += f" -hwaccel {self.api.lower()}"
        self.name = f'FFmpeg-{self.codec.value}{"-" + self.api if self.api else ""}'
        self.description = f'FFmpeg {self.codec.value} {self.api if self.hw_acceleration else "SW"} decoder'

    @lru_cache(maxsize=128)
    def ffmpeg_version(self) -> Optional[Match[str]]:
        """Returns the ffmpeg version as a re.Match object"""
        cmd = shlex.split("ffmpeg -version")
        output = subprocess.check_output(cmd, stderr=subprocess.DEVNULL).decode("utf-8")
        version = re.search(r"\d\.\d\.\d", output)
        return version

    def ffmpeg_cmd(
        self, input_filepath: str, output_filepath: str, output_format: OutputFormat
    ) -> List[str]:
        """Returns the formatted ffmpeg command based on the current ffmpeg version"""
        version = self.ffmpeg_version()
        download = ""
        if self.hw_acceleration and self.hw_download:
            download = f"hwdownload,format={output_format_to_ffformat (output_format)},"
        if version and int(version.group(0)[0]) >= 5 and int(version.group(0)[2]) >= 1:
            cmd = shlex.split(
                FFMPEG_TPL.format(
                    self.cmd,
                    input_filepath,
                    "-fps_mode passthrough",
                    download,
                    str(output_format.value),
                    output_filepath,
                )
            )
        else:
            cmd = shlex.split(
                FFMPEG_TPL.format(
                    self.cmd,
                    input_filepath,
                    "-vsync passthrough",
                    download,
                    str(output_format.value),
                    output_filepath,
                )
            )
        return cmd

    def decode(
        self,
        input_filepath: str,
        output_filepath: str,
        output_format: OutputFormat,
        timeout: int,
        verbose: bool,
        keep_files: bool,
    ) -> str:
        """Decodes input_filepath in output_filepath"""
        # pylint: disable=unused-argument
        cmd = self.ffmpeg_cmd(input_filepath, output_filepath, output_format)
        run_command(cmd, timeout=timeout, verbose=verbose)
        return file_checksum(output_filepath)

    @lru_cache(maxsize=128)
    def check(self, verbose: bool) -> bool:
        """Checks whether the decoder can be run"""
        # pylint: disable=broad-except
        if self.hw_acceleration:
            try:
                command = None

                if self.wrapper:
                    command = [self.binary, "-decoders"]
                else:
                    command = [self.binary, "-hwaccels"]

                output = subprocess.check_output(
                    command, stderr=subprocess.DEVNULL
                ).decode("utf-8")
                if verbose:
                    print(f'{" ".join(command)}\n{output}')

                if self.wrapper:
                    return self.api.lower() in output

                return f"{os.linesep}{self.api.lower()}{os.linesep}" in output
            except Exception:
                return False
        else:
            return super().check(verbose)


@register_decoder
class FFmpegH264Decoder(FFmpegDecoder):
    """FFmpeg SW decoder for H.264"""

    codec = Codec.H264


@register_decoder
class FFmpegH265Decoder(FFmpegDecoder):
    """FFmpeg SW decoder for H.265"""

    codec = Codec.H265


@register_decoder
class FFmpegVP8Decoder(FFmpegDecoder):
    """FFmpeg SW decoder for VP8"""

    codec = Codec.VP8


@register_decoder
class FFmpegVP9Decoder(FFmpegDecoder):
    """FFmpeg SW decoder for VP9"""

    codec = Codec.VP9


class FFmpegVaapiDecoder(FFmpegDecoder):
    """Generic class for FFmpeg VAAPI decoder"""

    hw_acceleration = True
    api = "VAAPI"


@register_decoder
class FFmpegH264VaapiDecoder(FFmpegVaapiDecoder):
    """FFmpeg VAAPI decoder for H.264"""

    codec = Codec.H264


@register_decoder
class FFmpegH265VaapiDecoder(FFmpegVaapiDecoder):
    """FFmpeg VAAPI decoder for H.265"""

    codec = Codec.H265


@register_decoder
class FFmpegVP8VaapiDecoder(FFmpegVaapiDecoder):
    """FFmpeg VAAPI decoder for VP8"""

    codec = Codec.VP8


@register_decoder
class FFmpegVP9VaapiDecoder(FFmpegVaapiDecoder):
    """FFmpeg VAAPI decoder for VP9"""

    codec = Codec.VP9


@register_decoder
class FFmpegAV1VaapiDecoder(FFmpegVaapiDecoder):
    """FFmpeg VAAPI decoder for AV1"""

    codec = Codec.AV1


class FFmpegVdpauDecoder(FFmpegDecoder):
    """Generic class for FFmpeg VDPAU decoder"""

    hw_acceleration = True
    api = "VDPAU"


@register_decoder
class FFmpegH264VdpauDecoder(FFmpegVdpauDecoder):
    """FFmpeg VDPAU decoder for H.264"""

    codec = Codec.H264


@register_decoder
class FFmpegH265VdpauDecoder(FFmpegVdpauDecoder):
    """FFmpeg VDPAU decoder for H.265"""

    codec = Codec.H265


class FFmpegDxva2Decoder(FFmpegDecoder):
    """Generic class for FFmpeg DXVA2 decoder"""

    hw_acceleration = True
    api = "DXVA2"


@register_decoder
class FFmpegH264Dxva2Decoder(FFmpegDxva2Decoder):
    """FFmpeg DXVA2 decoder for H.264"""

    codec = Codec.H264


@register_decoder
class FFmpegH265Dxva2Decoder(FFmpegDxva2Decoder):
    """FFmpeg DXVA2 decoder for H.265"""

    codec = Codec.H265


class FFmpegD3d11vaDecoder(FFmpegDecoder):
    """Generic class for FFmpeg D3D11VA decoder"""

    hw_acceleration = True
    api = "D3D11VA"


@register_decoder
class FFmpegH264D3d11vaDecoder(FFmpegD3d11vaDecoder):
    """FFmpeg D3D11VA decoder for H.264"""

    codec = Codec.H264


@register_decoder
class FFmpegH265D3d11vaDecoder(FFmpegD3d11vaDecoder):
    """FFmpeg D3D11VA decoder for H.265"""

    codec = Codec.H265


@register_decoder
class FFmpegVP8V4L2m2mDecoder(FFmpegDecoder):
    """FFmpeg V4L2m2m decoder for VP8"""

    codec = Codec.VP8
    hw_acceleration = True
    api = "vp8_v4l2m2m"
    wrapper = True


@register_decoder
class FFmpegVP9V4L2m2mDecoder(FFmpegDecoder):
    """FFmpeg V4L2m2m decoder for VP9"""

    codec = Codec.VP9
    hw_acceleration = True
    api = "vp9_v4l2m2m"
    wrapper = True


@register_decoder
class FFmpegH264V4L2m2mDecoder(FFmpegDecoder):
    """FFmpeg V4L2m2m decoder for H264"""

    codec = Codec.H264
    hw_acceleration = True
    api = "h264_v4l2m2m"
    wrapper = True


class FFmpegVulkanDecoder(FFmpegDecoder):
    """Generic class for FFmpeg Vulkan decoder"""

    hw_acceleration = True
    api = "Vulkan"
    init_device = "vulkan"
    hw_download = True


@register_decoder
class FFmpegH264VulkanDecoder(FFmpegVulkanDecoder):
    """FFmpeg Vulkan decoder for H.264"""

    codec = Codec.H264


@register_decoder
class FFmpegH265VulkanDecoder(FFmpegVulkanDecoder):
    """FFmpeg Vulkan decoder for H.265"""

    codec = Codec.H265


@register_decoder
class FFmpegAV1VulkanDecoder(FFmpegVulkanDecoder):
    """FFmpeg Vulkan decoder for AV1"""

    codec = Codec.AV1



# Fluster - testing framework for decoders conformance
# Copyright (C) 2020, Fluendo, S.A.
#  Author: Pablo Marcos Oltra <pmarcos@fluendo.com>, Fluendo, S.A.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <https://www.gnu.org/licenses/>.

from functools import lru_cache
import os.path
import json
import copy
import sys
from multiprocessing import Pool
from unittest.result import TestResult
from time import perf_counter
from shutil import rmtree
from typing import cast, List, Dict, Optional, Type, Any
import urllib.error


from fluster.test_vector import TestVector
from fluster.codec import Codec
from fluster.decoder import Decoder
from fluster.test import Test
from fluster import utils


class DownloadWork:
    """Context to pass to each download worker"""

    def __init__(
        self,
        out_dir: str,
        verify: bool,
        extract_all: bool,
        keep_file: bool,
        test_suite_name: str,
        test_vector: TestVector,
        retries: int,
    ):
        self.out_dir = out_dir
        self.verify = verify
        self.extract_all = extract_all
        self.keep_file = keep_file
        self.test_suite_name = test_suite_name
        self.test_vector = test_vector
        self.retries = retries


class Context:
    """Context for TestSuite"""

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        jobs: int,
        decoder: Decoder,
        timeout: int,
        failfast: bool,
        quiet: bool,
        output_dir: str,
        reference: bool = False,
        test_vectors: Optional[List[str]] = None,
        skip_vectors: Optional[List[str]] = None,
        failing_test_vectors: Optional[List[str]] = None,
        keep_files: bool = False,
        verbose: bool = False,
    ):
        self.jobs = jobs
        self.decoder = decoder
        self.timeout = timeout
        self.failfast = failfast
        self.quiet = quiet
        self.output_dir = output_dir
        self.reference = reference
        self.test_vectors = test_vectors
        self.skip_vectors = skip_vectors
        self.failing_test_vectors = failing_test_vectors
        self.keep_files = keep_files
        self.verbose = verbose


class TestSuite:
    """Test suite class"""

    # pylint: disable=too-many-instance-attributes

    TEST_SUITE_NAME = "TEST SUITE"
    DECODER_NAME = "DECODER"
    TEST_VECTOR_NAME = "TEST VECTOR"
    RESULT_NAME = "RESULT"

    def __init__(
        self,
        filename: str,
        resources_dir: str,
        name: str,
        codec: Codec,
        description: str,
        test_vectors: Dict[str, TestVector],
        failing_test_vectors: Optional[Dict[str, TestVector]] = None,
    ):
        # JSON members
        self.name = name
        self.codec = codec
        self.description = description
        self.test_vectors = test_vectors
        self.failing_test_vectors = failing_test_vectors

        # Not included in JSON
        self.filename = filename
        self.resources_dir = resources_dir
        self.test_vectors_success = 0
        self.time_taken = 0.0

    def clone(self) -> "TestSuite":
        """Create a deep copy of the object"""
        return copy.deepcopy(self)

    @classmethod
    def from_json_file(
        cls: Type["TestSuite"], filename: str, resources_dir: str
    ) -> "TestSuite":
        """Create a TestSuite instance from a file"""
        with open(filename, encoding="utf-8") as json_file:
            data = json.load(json_file)
            if "failing_test_vectors" in data:
                data["failing_test_vectors"] = dict(
                    map(TestVector.from_json, data["failing_test_vectors"])
                )
            data["test_vectors"] = dict(map(TestVector.from_json, data["test_vectors"]))
            data["codec"] = Codec(data["codec"])
            return cls(filename, resources_dir, **data)

    def to_json_file(self, filename: str) -> None:
        """Serialize the test suite to a file"""
        with open(filename, "w", encoding="utf-8") as json_file:
            data = self.__dict__.copy()
            data.pop("resources_dir")
            data.pop("filename")
            data.pop("test_vectors_success")
            data.pop("time_taken")
            if self.failing_test_vectors is None:
                data.pop("failing_test_vectors")
            else:
                data["failing_test_vectors"] = [
                    failing_test_vector.data_to_serialize()
                    for failing_test_vector in self.failing_test_vectors.values()
                ]
            data["codec"] = str(self.codec.value)
            data["test_vectors"] = [
                test_vector.data_to_serialize()
                for test_vector in self.test_vectors.values()
            ]
            json.dump(data, json_file, indent=4)

    def _download_worker(self, ctx: DownloadWork) -> None:
        """Download and extract a test vector"""
        test_vector = ctx.test_vector
        dest_dir = os.path.join(ctx.out_dir, ctx.test_suite_name, test_vector.name)
        dest_path = os.path.join(dest_dir, os.path.basename(test_vector.source))
        if not os.path.exists(dest_dir):
            os.makedirs(dest_dir)
        if (
            ctx.verify
            and os.path.exists(dest_path)
            and test_vector.source_checksum == utils.file_checksum(dest_path)
        ):
            # Remove file only in case the input file was extractable.
            # Otherwise, we'd be removing the original file we want to work
            # with every even time we execute the download subcommand.
            if utils.is_extractable(dest_path) and not ctx.keep_file:
                os.remove(dest_path)
            return
        print(f"\tDownloading test vector {test_vector.name} from {dest_dir}")
        # Catch the exception that download may throw to make sure pickle can serialize it properly
        # This avoids:
        # Error sending result: '<multiprocessing.pool.ExceptionWithTraceback object at 0x7fd7811ecee0>'.
        # Reason: 'TypeError("cannot pickle '_io.BufferedReader' object")'
        for i in range(ctx.retries):
            try:
                exception_str = ""
                utils.download(test_vector.source, dest_dir)
            except urllib.error.URLError as ex:
                exception_str = str(ex)
                print(
                    f"\tUnable to download {test_vector.source} to {dest_dir}, {exception_str}, retry count={i+1}"
                )
                continue
            except Exception as ex:
                raise Exception(str(ex)) from ex
            break

        if exception_str:
            raise Exception(exception_str)

        if test_vector.source_checksum != "__skip__":
            checksum = utils.file_checksum(dest_path)
            if test_vector.source_checksum != checksum:
                raise Exception(
                    f"Checksum error for test vector '{test_vector.name}': '{checksum}' instead of "
                    f"'{test_vector.source_checksum}'"
                )

        if utils.is_extractable(dest_path):
            print(f"\tExtracting test vector {test_vector.name} to {dest_dir}")
            utils.extract(
                dest_path,
                dest_dir,
                file=test_vector.input_file if not ctx.extract_all else None,
            )
            if not ctx.keep_file:
                os.remove(dest_path)

    def download(
        self,
        jobs: int,
        out_dir: str,
        verify: bool,
        extract_all: bool = False,
        keep_file: bool = False,
        retries: int = 1,
    ) -> None:
        """Download the test suite"""
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        print(f"Downloading test suite {self.name} using {jobs} parallel jobs")

        with Pool(jobs) as pool:

            def _callback_error(err: Any) -> None:
                print(f"\nError downloading -> {err}\n")
                pool.terminate()

            downloads = []
            for test_vector in self.test_vectors.values():
                dwork = DownloadWork(
                    out_dir,
                    verify,
                    extract_all,
                    keep_file,
                    self.name,
                    test_vector,
                    retries,
                )
                downloads.append(
                    pool.apply_async(
                        self._download_worker,
                        args=(dwork,),
                        error_callback=_callback_error,
                    )
                )
            pool.close()
            pool.join()

        for job in downloads:
            if not job.successful():
                sys.exit("Some download failed")

        print("All downloads finished")

    def _rename_test(self, test: Test, module: str, qualname: str) -> None:
        test_cls = type(test)
        test_cls.__module__ = module
        test_cls.__qualname__ = qualname

    def _collect_results(self, test_result: TestResult) -> None:
        """Collect all TestResults with error to add them into the test vectors"""
        for res in test_result.failures:
            test_vector = cast(Test, res[0]).test_vector
            test_vector.errors.append([str(x) for x in res])
        for res in test_result.errors:
            test_vector = cast(Test, res[0]).test_vector
            test_vector.errors.append([str(x) for x in res])

    def _run_worker(self, test: Test) -> TestVector:
        """Run one unit test returning the TestVector"""
        # Save the original module and qualname to restore it before returning
        # the TestVector. Otherwise, Pickle will complain if the classes can't
        # be found in global scope. The trick here is that we change the names
        # momentarily just to obtain the error traces in str format
        test_cls = type(test)
        module_orig = test_cls.__module__
        qualname_orig = test_cls.__qualname__
        self._rename_test(test, test.decoder.name, test.test_suite.name)

        test_result = TestResult()
        test(test_result)

        self._collect_results(test_result)
        self._rename_test(test, module_orig, qualname_orig)

        return test.test_vector

    def _get_max_length_list_name(self, _list: List[str], name: str) -> int:
        max_length = len(name)
        for elem in _list:
            length = len(elem)
            max_length = max(max_length, length)
        return max_length

    @lru_cache(maxsize=128)
    def _get_result_line(
        self,
        test_suite_text: str,
        decoder_text: str,
        test_vector_text: str,
        result_text: str,
        decoder_name: Optional[str] = None,
    ) -> str:
        decoder_name = decoder_text if not decoder_name else decoder_name
        tests_suite_max_len = self._get_max_length_list_name(
            [self.name], TestSuite.TEST_SUITE_NAME
        )
        decoder_max_len = self._get_max_length_list_name(
            [decoder_name], TestSuite.DECODER_NAME
        )
        test_vectors_max_len = self._get_max_length_list_name(
            list(self.test_vectors.keys()), TestSuite.TEST_VECTOR_NAME
        )

        return (
            f"[{test_suite_text:{tests_suite_max_len}}] ({decoder_text:{decoder_max_len}}) "
            f"{test_vector_text:{test_vectors_max_len}} ... {result_text}"
        )

    def run_test_suite_in_parallel(
        self, jobs: int, tests: List[Test], failfast: bool
    ) -> None:
        """Run the test suite in parallel"""
        test_vector_results: List[TestVector] = []
        decoder = tests[0].decoder

        print(
            self._get_result_line(
                TestSuite.TEST_SUITE_NAME,
                TestSuite.DECODER_NAME,
                TestSuite.TEST_VECTOR_NAME,
                TestSuite.RESULT_NAME,
                decoder.name,
            )
            + f'\n{"-" * 70}'
        )
        with Pool(jobs) as pool:

            def _callback(test_result: TestVector) -> None:
                print(
                    self._get_result_line(
                        self.name,
                        decoder.name,
                        test_result.name,
                        test_result.test_result.value,
                    ),
                    flush=True,
                )
                test_vector_results.append(test_result)
                if failfast and test_result.errors:
                    pool.terminate()

            start = perf_counter()
            for test in tests:
                pool.apply_async(self._run_worker, (test,), callback=_callback)
            pool.close()
            pool.join()
        self.time_taken = perf_counter() - start
        print("\n")
        self.test_vectors_success = 0
        for test_vector_res in test_vector_results:
            if test_vector_res.errors:
                for error in test_vector_res.errors:
                    # Use same format to report errors as TextTestRunner
                    print(f'{"=" * 71}\nFAIL: {error[0]}\n{"-" * 70}')
                    for line in error[1:]:
                        print(line)
            else:
                self.test_vectors_success += 1

            # Collect the test vector results and failures since they come
            # from a different process
            self.test_vectors[test_vector_res.name] = test_vector_res
        print(
            f"Ran {self.test_vectors_success}/{len(tests)} tests successfully \
              in {self.time_taken:.3f} secs"
        )

    def run(self, ctx: Context) -> Optional["TestSuite"]:
        """
        Run the test suite.
        Returns a new copy of the test suite with the result of the test
        """
        # pylint: disable=too-many-locals

        if not ctx.decoder.check(ctx.verbose):
            print(f"Skipping decoder {ctx.decoder.name} because it cannot be run")
            return None

        ctx.output_dir = os.path.join(ctx.output_dir, self.name)
        if os.path.exists(ctx.output_dir):
            rmtree(ctx.output_dir)
        os.makedirs(ctx.output_dir)

        test_suite = self.clone()
        tests = test_suite.generate_tests(ctx)
        if not tests:
            extra = ""
            if ctx.test_vectors:
                extra = "with names: " + ", ".join(ctx.test_vectors)
            print(f"No test vectors for suite {self.name} {extra}")
            return None

        print("*" * 100)
        string = f"Running test suite {self.name} with decoder {ctx.decoder.name}\n"
        if ctx.test_vectors:
            string += f'Test vectors {" ".join(ctx.test_vectors)}\n'
        if ctx.skip_vectors:
            string += f'Skipping test vectors {" ".join(ctx.skip_vectors)}\n'
        string += f"Using {ctx.jobs} parallel job(s)"
        print(string)
        print("*" * 100 + "\n")

        test_suite.run_test_suite_in_parallel(ctx.jobs, tests, ctx.failfast)

        if ctx.reference:
            test_suite.to_json_file(test_suite.filename)

        if not ctx.keep_files and os.path.isdir(ctx.output_dir):
            rmtree(ctx.output_dir)

        return test_suite

    def generate_tests(self, ctx: Context) -> List[Test]:
        """Generate the tests for a decoder"""
        tests = []
        test_vectors_run = {}
        for name, test_vector in self.test_vectors.items():
            skip = False
            if ctx.test_vectors:
                if test_vector.name.lower() not in ctx.test_vectors:
                    continue
            if ctx.skip_vectors:
                if test_vector.name.lower() in ctx.skip_vectors:
                    skip = True
            tests.append(
                Test(
                    ctx.decoder,
                    self,
                    test_vector,
                    skip,
                    ctx.output_dir,
                    ctx.reference,
                    ctx.timeout,
                    ctx.keep_files,
                    ctx.verbose,
                )
            )
            test_vectors_run[name] = test_vector
        self.test_vectors = test_vectors_run
        return tests

    def __str__(self) -> str:
        return (
            f"\n{self.name}\n"
            f"    Codec: {self.codec.value}\n"
            f"    Description: {self.description}\n"
            f"    Test vectors: {len(self.test_vectors)}"
        )



# Fluster - testing framework for decoders conformance
# Copyright (C) 2020, Fluendo, S.A.
#  Author: Pablo Marcos Oltra <pmarcos@fluendo.com>, Fluendo, S.A.
#  Author: Andoni Morales Alastruey <amorales@fluendo.com>, Fluendo, S.A.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <https://www.gnu.org/licenses/>.

import hashlib
import os
import shutil
import subprocess
import sys
import urllib.request
import zipfile
import platform
from typing import List, Optional


TARBALL_EXTS = ("tar.gz", "tgz", "tar.bz2", "tbz2", "tar.xz")


def download(url: str, dest_dir: str) -> None:
    """Downloads a file to a directory"""
    with urllib.request.urlopen(url) as response:
        dest_path = os.path.join(dest_dir, url.split("/")[-1])
        with open(dest_path, "wb") as dest:
            shutil.copyfileobj(response, dest)


def file_checksum(path: str) -> str:
    """Calculates the checksum of a file reading chunks of 64KiB"""
    md5 = hashlib.md5()
    with open(path, "rb") as file:
        while True:
            data = file.read(65536)
            if not data:
                break
            md5.update(data)
    return md5.hexdigest()


def run_command(
    command: List[str],
    verbose: bool = False,
    check: bool = True,
    timeout: Optional[int] = None,
) -> None:
    """Runs a command"""
    sout = subprocess.DEVNULL if not verbose else None
    serr = subprocess.DEVNULL if not verbose else None
    if verbose:
        print(f'\nRunning command "{" ".join(command)}"')
    try:
        subprocess.run(command, stdout=sout, stderr=serr, check=check, timeout=timeout)
    except (subprocess.CalledProcessError, subprocess.TimeoutExpired) as ex:
        # Developer experience improvement (facilitates copy/paste)
        ex.cmd = " ".join(ex.cmd)
        raise ex


def run_pipe_command_with_std_output(
    command: List[str],
    verbose: bool = False,
    check: bool = True,
    timeout: Optional[int] = None,
) -> List[str]:
    """Runs a command and returns std output trace"""
    serr = subprocess.DEVNULL if not verbose else subprocess.STDOUT
    if verbose:
        print(f'\nRunning command "{" ".join(command)}"')

    try:
        data = subprocess.check_output(
            command, stderr=serr, timeout=timeout, universal_newlines=True
        )
        return data.splitlines()
    except (subprocess.CalledProcessError, subprocess.TimeoutExpired) as ex:
        odata: List[str] = []
        if verbose or check:
            # Workaround inconsistent Python implementation
            if isinstance(ex, subprocess.CalledProcessError):
                odata = ex.output.splitlines()
            else:
                odata = ex.output.decode("utf-8").splitlines()
        if verbose:
            for line in odata:
                print(line)

        if isinstance(ex, subprocess.CalledProcessError) and not check:
            return odata

        # Developer experience improvement (facilitates copy/paste)
        ex.cmd = " ".join(ex.cmd)
        raise ex


def is_extractable(filepath: str) -> bool:
    """Checks is a file can be extracted from the its extension"""
    return filepath.endswith(TARBALL_EXTS) or filepath.endswith(".zip")


def extract(filepath: str, output_dir: str, file: Optional[str] = None) -> None:
    """Extracts a file to a directory"""
    if filepath.endswith(TARBALL_EXTS):
        command = ["tar", "-C", output_dir, "-xf", filepath]
        if file:
            command.append(file)
        subprocess.run(command, check=True)
    elif filepath.endswith(".zip"):
        with zipfile.ZipFile(filepath, "r") as zip_file:
            if file:
                zip_file.extract(file, path=output_dir)
            else:
                zip_file.extractall(path=output_dir)
    else:
        raise Exception(f"Unknown tarball format {filepath}")


def normalize_binary_cmd(cmd: str) -> str:
    """Return the OS-form binary"""
    if platform.system() == "Windows":
        return cmd if cmd.endswith(".exe") else cmd + ".exe"
    if cmd.endswith(".exe"):
        return cmd.replace(".exe", "")
    return cmd


def normalize_path(path: str) -> str:
    """Normalize the path to make it Unix-like"""
    if platform.system() == "Windows":
        return path.replace("\\", "/")
    return path


def _linux_user_data_dir(appname: str) -> str:
    """Return data directory tied to the user"""
    path = os.environ.get("XDG_DATA_HOME", "")
    if not path.strip():
        path = os.path.expanduser("~/.local/share")
    return os.path.join(path, appname)


def _linux_site_data_dirs(appname: str) -> List[str]:
    """Return data directory shared by users"""
    path = os.environ.get("XDG_DATA_DIRS", "")
    if not path.strip():
        path = "/usr/local/share:/usr/share"
    paths = path.split(os.pathsep)
    return [os.path.join(p, appname) for p in paths]


def _win_user_data_dir(appname: str) -> str:
    """Return data directory"""
    path = os.path.expanduser(r"~\AppData\Local")
    return os.path.join(path, appname)


def _win_site_data_dirs(appname: str) -> List[str]:
    """Return data directory shared by users"""
    # On Windows always user_data_dir
    return [_win_user_data_dir(appname)]


if sys.platform == "win32":
    site_data_dirs = _win_site_data_dirs
    user_data_dir = _win_user_data_dir
else:
    site_data_dirs = _linux_site_data_dirs
    user_data_dir = _linux_user_data_dir



# Fluster - testing framework for decoders conformance
# Copyright (C) 2020, Fluendo, S.A.
#  Author: Pablo Marcos Oltra <pmarcos@fluendo.com>, Fluendo, S.A.
#  Author: Andoni Morales Alastruey <amorales@fluendo.com>, Fluendo, S.A.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <https://www.gnu.org/licenses/>.

import shlex
import subprocess
from functools import lru_cache
from typing import Optional, List

from fluster.codec import Codec, OutputFormat
from fluster.decoder import Decoder, register_decoder
from fluster.utils import (
    file_checksum,
    run_command,
    run_pipe_command_with_std_output,
    normalize_binary_cmd,
)

PIPELINE_TPL = "{} --no-fault filesrc location={} ! {} ! {} ! {} {}"


@lru_cache(maxsize=None)
def gst_element_exists(element: str) -> bool:
    """Check if an element exists in current GStreamer installation"""
    inspect_exe = normalize_binary_cmd("gst-inspect-1.0")

    try:
        subprocess.run(
            [inspect_exe, "--exists", element],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            check=True,
        )
        return True
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False
    return False


def output_format_to_gst(output_format: OutputFormat) -> str:
    """Return GStreamer pixel format"""
    mapping = {
        OutputFormat.GRAY: "GRAY8",
        OutputFormat.GRAY16LE: "GRAY16_LE",
        OutputFormat.YUV420P: "I420",
        OutputFormat.YUV422P: "Y42B",
        OutputFormat.YUV420P10LE: "I420_10LE",
        OutputFormat.YUV422P10LE: "I422_10LE",
        OutputFormat.YUV420P12LE: "I420_12LE",
        OutputFormat.YUV422P12LE: "I422_12LE",
        OutputFormat.YUV444P: "Y444",
        OutputFormat.YUV444P10LE: "Y444_10LE",
        OutputFormat.YUV444P12LE: "Y444_12LE",
    }
    if output_format not in mapping:
        raise Exception(
            f"No matching output format found in GStreamer for {output_format}"
        )
    return mapping[output_format]


class GStreamer(Decoder):
    """Base class for GStreamer decoders"""

    decoder_bin = ""
    cmd = ""
    caps = ""
    gst_api = ""
    api = ""
    provider = ""
    sink = ""

    def __init__(self) -> None:
        super().__init__()
        if not self.name:
            self.name = (
                f"{self.provider}-{self.codec.value}-{self.api}-Gst{self.gst_api}"
            )
        self.description = f"{self.provider} {self.codec.value} {self.api} decoder for GStreamer {self.gst_api}"
        self.cmd = normalize_binary_cmd(self.cmd)

        if not gst_element_exists(self.sink):
            self.sink = "filesink"

    def gen_pipeline(
        self,
        input_filepath: str,
        output_filepath: Optional[str],
        output_format: OutputFormat,
    ) -> str:
        """Generate the GStreamer pipeline used to decode the test vector"""
        # pylint: disable=unused-argument
        output = f"location={output_filepath}" if output_filepath else ""
        return PIPELINE_TPL.format(
            self.cmd, input_filepath, self.decoder_bin, self.caps, self.sink, output
        )

    def parse_videocodectestsink_md5sum(self, data: List[str], verbose: bool) -> str:
        """Parse the MD5 sum out of commandline output produced when using
        videocodectestsink."""
        md5sum = None
        for line in data:
            if verbose:
                print(line)
            pattern = (
                "conformance/checksum, checksum-type=(string)MD5, checksum=(string)"
            )
            sum_start = line.find(pattern)
            # pylint: disable=no-else-continue
            if sum_start <= 0:
                # Skip to the next iteration if sum_start is less than or equal to 0
                continue
            else:
                sum_start += len(pattern)
                sum_end = line[sum_start:].find(";")
                # pylint: disable=no-else-continue
                if sum_end <= 0:
                    # Skip to the next iteration if sum_end is less than or equal to 0
                    continue
                else:
                    sum_end += sum_start
                    md5sum = line[sum_start:sum_end]
                    if not verbose:
                        return md5sum

        if not md5sum:
            raise Exception("No MD5 found in the program trace.")

        return md5sum

    def decode(
        self,
        input_filepath: str,
        output_filepath: str,
        output_format: OutputFormat,
        timeout: int,
        verbose: bool,
        keep_files: bool,
    ) -> str:
        """Decode the test vector and do the checksum"""
        # When using videocodectestsink we can avoid writing files to disk
        # completely, or avoid a full raw file read in order to compute the MD5
        # SUM.
        if self.sink == "videocodectestsink":
            output_param = output_filepath if keep_files else None
            pipeline = self.gen_pipeline(input_filepath, output_param, output_format)
            command = shlex.split(pipeline)
            command.append("-m")
            data = run_pipe_command_with_std_output(
                command, timeout=timeout, verbose=verbose
            )
            return self.parse_videocodectestsink_md5sum(data, verbose)

        pipeline = self.gen_pipeline(input_filepath, output_filepath, output_format)
        run_command(shlex.split(pipeline), timeout=timeout, verbose=verbose)
        return file_checksum(output_filepath)

    @lru_cache(maxsize=128)
    def check(self, verbose: bool) -> bool:
        """Check if GStreamer decoder is valid (better than gst-inspect)"""
        # pylint: disable=broad-except
        try:
            if hasattr(self, "check_decoder_bin"):
                decoder_bin = getattr(self, "check_decoder_bin")
            else:
                decoder_bin = self.decoder_bin
            binary = normalize_binary_cmd(f"gst-launch-{self.gst_api}")
            pipeline = (
                f"{binary} --no-fault appsrc num-buffers=0 ! {decoder_bin} ! fakesink"
            )
            run_command(shlex.split(pipeline), verbose=verbose)
        except Exception:
            return False
        return True


class GStreamer10Video(GStreamer):
    """Base class for GStreamer 1.x video decoders"""

    cmd = "gst-launch-1.0"
    caps = "video/x-raw"
    gst_api = "1.0"
    sink = "videocodectestsink"
    provider = "GStreamer"

    def gen_pipeline(
        self,
        input_filepath: str,
        output_filepath: Optional[str],
        output_format: OutputFormat,
    ) -> str:
        caps = f"{self.caps} ! videoconvert dither=none ! video/x-raw,format={output_format_to_gst(output_format)}"
        output = f"location={output_filepath}" if output_filepath else ""
        return PIPELINE_TPL.format(
            self.cmd, input_filepath, self.decoder_bin, caps, self.sink, output
        )


class GStreamer10Audio(GStreamer):
    """Base class for GStreamer 1.x audio decoders"""

    cmd = "gst-launch-1.0"
    caps = "audio/x-raw"
    gst_api = "1.0"
    sink = "filesink"
    provider = "GStreamer"


class GStreamer010Video(GStreamer):
    """Base class for GStreamer 0.10 video decoders"""

    cmd = "gst-launch-0.10"
    caps = "video/x-raw-yuv"
    gst_api = "0.10"
    sink = "videocodectestsink"
    provider = "GStreamer"


@register_decoder
class GStreamerLibavH264(GStreamer10Video):
    """GStreamer H.264 Libav decoder implementation for GStreamer 1.0"""

    codec = Codec.H264
    decoder_bin = " h264parse ! avdec_h264 "
    api = "Libav"
    hw_acceleration = False


@register_decoder
class GStreamerLibavH265(GStreamer10Video):
    """GStreamer H.265 Libav decoder implementation for GStreamer 1.0"""

    codec = Codec.H265
    decoder_bin = " h265parse ! avdec_h265 "
    api = "Libav"
    hw_acceleration = False


@register_decoder
class GStreamerLibavVP8(GStreamer10Video):
    """GStreamer VP8 Libav decoder implementation for GStreamer 1.0"""

    codec = Codec.VP8
    decoder_bin = " ivfparse ! avdec_vp8 "
    api = "Libav"
    hw_acceleration = False


@register_decoder
class GStreamerLibavVP9(GStreamer10Video):
    """GStreamer VP9 Libav decoder implementation for GStreamer 1.0"""

    codec = Codec.VP9
    check_decoder_bin = " avdec_vp9"
    decoder_bin = f" parsebin ! {check_decoder_bin}"
    api = "Libav"
    hw_acceleration = False


@register_decoder
class GStreamerVaapiH265Gst10Decoder(GStreamer10Video):
    """GStreamer H.265 VAAPI decoder implementation for GStreamer 1.0"""

    codec = Codec.H265
    decoder_bin = " h265parse ! vaapih265dec "
    api = "VAAPI"
    hw_acceleration = True


@register_decoder
class GStreamerVaH265Gst10Decoder(GStreamer10Video):
    """GStreamer H.265 VA decoder implementation for GStreamer 1.0"""

    codec = Codec.H265
    decoder_bin = " h265parse ! vah265dec "
    api = "VA"
    hw_acceleration = True


@register_decoder
class GStreamerMsdkH265Gst10Decoder(GStreamer10Video):
    """GStreamer H.265 Intel MSDK decoder implementation for GStreamer 1.0"""

    codec = Codec.H265
    decoder_bin = " h265parse ! msdkh265dec "
    api = "MSDK"
    hw_acceleration = True


@register_decoder
class GStreamerNvdecH265Gst10Decoder(GStreamer10Video):
    """GStreamer H.265 NVDEC decoder implementation for GStreamer 1.0"""

    codec = Codec.H265
    decoder_bin = " h265parse ! nvh265dec "
    api = "NVDEC"
    hw_acceleration = True


@register_decoder
class GStreamerNvdecSLH265Gst10Decoder(GStreamer10Video):
    """GStreamer H.265 NVDEC stateless decoder implementation for GStreamer 1.0"""

    codec = Codec.H265
    decoder_bin = " h265parse ! nvh265sldec "
    api = "NVDECSL"
    hw_acceleration = True


@register_decoder
class GStreamerD3d11H265Gst10Decoder(GStreamer10Video):
    """GStreamer H.265 D3D11 decoder implementation for GStreamer 1.0"""

    codec = Codec.H265
    decoder_bin = " h265parse ! d3d11h265dec "
    api = "D3D11"
    hw_acceleration = True


@register_decoder
class GStreamerD3d12H265Gst10Decoder(GStreamer10Video):
    """GStreamer H.265 D3D12 decoder implementation for GStreamer 1.0"""

    codec = Codec.H265
    decoder_bin = " h265parse ! d3d12h265dec "
    api = "D3D12"
    hw_acceleration = True


@register_decoder
class GStreamerV4l2CodecsH265Gst10Decoder(GStreamer10Video):
    """GStreamer H.265 V4L2 stateless decoder implementation for GStreamer 1.0"""

    codec = Codec.H265
    decoder_bin = " h265parse ! v4l2slh265dec "
    api = "V4L2SL"
    hw_acceleration = True


@register_decoder
class GStreamerV4l2H265Gst10Decoder(GStreamer10Video):
    """GStreamer H.265 V4L2 stateful decoder implementation for GStreamer 1.0"""

    codec = Codec.H265
    decoder_bin = " h265parse ! v4l2h265dec "
    api = "V4L2"
    hw_acceleration = True


@register_decoder
class GStreamerVaapiH264Gst10Decoder(GStreamer10Video):
    """GStreamer H.264 VAAPI decoder implementation for GStreamer 1.0"""

    codec = Codec.H264
    decoder_bin = " h264parse ! vaapih264dec "
    api = "VAAPI"
    hw_acceleration = True


@register_decoder
class GStreamerVaH264Gst10Decoder(GStreamer10Video):
    """GStreamer H.264 VA decoder implementation for GStreamer 1.0"""

    codec = Codec.H264
    decoder_bin = " h264parse ! vah264dec "
    api = "VA"
    hw_acceleration = True


@register_decoder
class GStreamerMsdkH264Gst10Decoder(GStreamer10Video):
    """GStreamer H.264 Intel MSDK decoder implementation for GStreamer 1.0"""

    codec = Codec.H264
    decoder_bin = " h264parse ! msdkh264dec "
    api = "MSDK"
    hw_acceleration = True


@register_decoder
class GStreamerNvdecH264Gst10Decoder(GStreamer10Video):
    """GStreamer H.264 NVDEC decoder implementation for GStreamer 1.0"""

    codec = Codec.H264
    decoder_bin = " h264parse ! nvh264dec "
    api = "NVDEC"
    hw_acceleration = True


@register_decoder
class GStreamerNvdecSLH264Gst10Decoder(GStreamer10Video):
    """GStreamer H.264 NVDEC stateless decoder implementation for GStreamer 1.0"""

    codec = Codec.H264
    decoder_bin = " h264parse ! nvh264sldec "
    api = "NVDECSL"
    hw_acceleration = True


@register_decoder
class GStreamerD3d11H264Gst10Decoder(GStreamer10Video):
    """GStreamer H.264 D3D11 decoder implementation for GStreamer 1.0"""

    codec = Codec.H264
    decoder_bin = " h264parse ! d3d11h264dec "
    api = "D3D11"
    hw_acceleration = True


@register_decoder
class GStreamerD3d12H264Gst10Decoder(GStreamer10Video):
    """GStreamer H.264 D3D12 decoder implementation for GStreamer 1.0"""

    codec = Codec.H264
    decoder_bin = " h264parse ! d3d12h264dec "
    api = "D3D12"
    hw_acceleration = True


@register_decoder
class GStreamerV4l2CodecsH264Gst10Decoder(GStreamer10Video):
    """GStreamer H.264 V4L2 stateless decoder implementation for GStreamer 1.0"""

    codec = Codec.H264
    decoder_bin = " h264parse ! v4l2slh264dec "
    api = "V4L2SL"
    hw_acceleration = True


@register_decoder
class GStreamerV4l2H264Gst10Decoder(GStreamer10Video):
    """GStreamer H.264 V4L2 stateful decoder implementation for GStreamer 1.0"""

    codec = Codec.H264
    decoder_bin = " h264parse ! v4l2h264dec "
    api = "V4L2"
    hw_acceleration = True


@register_decoder
class GStreamerVulkanH264Gst10Decoder(GStreamer10Video):
    """GStreamer H.264 Vulkan stateless decoder implementation for GStreamer 1.0"""

    codec = Codec.H264
    decoder_bin = " h264parse ! vulkanh264dec ! vulkandownload "
    api = "Vulkan"
    hw_acceleration = True


@register_decoder
class GStreamerV4l2CodecsVP8Gst10Decoder(GStreamer10Video):
    """GStreamer VP8 V4L2 stateless decoder implementation for GStreamer 1.0"""

    codec = Codec.VP8
    decoder_bin = " ivfparse ! v4l2slvp8dec "
    api = "V4L2SL"
    hw_acceleration = True


@register_decoder
class GStreamerV4l2VP8Gst10Decoder(GStreamer10Video):
    """GStreamer VP8 V4L2 stateful decoder implementation for GStreamer 1.0"""

    codec = Codec.VP8
    decoder_bin = " ivfparse ! v4l2vp8dec "
    api = "V4L2"
    hw_acceleration = True


@register_decoder
class GStreamerLibvpxVP8(GStreamer10Video):
    """GStreamer VP8 Libvpx decoder implementation for GStreamer 1.0"""

    codec = Codec.VP8
    decoder_bin = " ivfparse ! vp8dec "
    api = "libvpx"
    hw_acceleration = False


@register_decoder
class GStreamerVaapiVP8Gst10Decoder(GStreamer10Video):
    """GStreamer VP8 VAAPI decoder implementation for GStreamer 1.0"""

    codec = Codec.VP8
    decoder_bin = " ivfparse ! vaapivp8dec "
    api = "VAAPI"
    hw_acceleration = True


@register_decoder
class GStreamerVaVP8Gst10Decoder(GStreamer10Video):
    """GStreamer VP8 VA decoder implementation for GStreamer 1.0"""

    codec = Codec.VP8
    decoder_bin = " ivfparse ! vavp8dec "
    api = "VA"
    hw_acceleration = True


@register_decoder
class GStreamerD3d11VP8Gst10Decoder(GStreamer10Video):
    """GStreamer VP8 D3D11 decoder implementation for GStreamer 1.0"""

    codec = Codec.VP8
    decoder_bin = " ivfparse ! d3d11vp8dec "
    api = "D3D11"
    hw_acceleration = True


@register_decoder
class GStreamerNvdecVP8Gst10Decoder(GStreamer10Video):
    """GStreamer VP8 NVDEC decoder implementation for GStreamer 1.0"""

    codec = Codec.VP8
    decoder_bin = " ivfparse ! nvvp8dec "
    api = "NVDEC"
    hw_acceleration = True


@register_decoder
class GStreamerNvdecSLVP8Gst10Decoder(GStreamer10Video):
    """GStreamer VP8 NVDEC stateless decoder implementation for GStreamer 1.0"""

    codec = Codec.VP8
    decoder_bin = " ivfparse ! nvvp8sldec "
    api = "NVDECSL"
    hw_acceleration = True


@register_decoder
class GStreamerV4l2CodecsVP9Gst10Decoder(GStreamer10Video):
    """GStreamer VP9 V4L2 stateless decoder implementation for GStreamer 1.0"""

    codec = Codec.VP9
    check_decoder_bin = " v4l2slvp9dec "
    decoder_bin = f" parsebin ! {check_decoder_bin}"
    api = "V4L2SL"
    hw_acceleration = True


@register_decoder
class GStreamerV4l2VP9Gst10Decoder(GStreamer10Video):
    """GStreamer VP9 V4L2 stateful decoder implementation for GStreamer 1.0"""

    codec = Codec.VP9
    check_decoder_bin = " v4l2vp9dec "
    decoder_bin = f" parsebin ! {check_decoder_bin}"
    api = "V4L2"
    hw_acceleration = True


@register_decoder
class GStreamerVaAV1Gst10Decoder(GStreamer10Video):
    """GStreamer AV1 VA decoder implementation for GStreamer 1.0"""

    codec = Codec.AV1
    check_decoder_bin = " vaav1dec "
    decoder_bin = f" parsebin ! {check_decoder_bin}"
    api = "VA"
    hw_acceleration = True


@register_decoder
class GStreamerV4l2CodecsAV1Gst10Decoder(GStreamer10Video):
    """GStreamer AV1 V4L2 stateless decoder implementation for GStreamer 1.0"""

    codec = Codec.AV1
    check_decoder_bin = " v4l2slav1dec "
    decoder_bin = f" parsebin ! {check_decoder_bin}"
    api = "V4L2SL"
    hw_acceleration = True


@register_decoder
class GStreamerD3d11AV1Gst10Decoder(GStreamer10Video):
    """GStreamer AV1 D3D11 decoder implementation for GStreamer 1.0"""

    codec = Codec.AV1
    check_decoder_bin = " d3d11av1dec "
    decoder_bin = f" parsebin ! {check_decoder_bin}"
    api = "D3D11"
    hw_acceleration = True


@register_decoder
class GStreamerD3d12AV1Gst10Decoder(GStreamer10Video):
    """GStreamer AV1 D3D12 decoder implementation for GStreamer 1.0"""

    codec = Codec.AV1
    check_decoder_bin = " d3d12av1dec "
    decoder_bin = f" parsebin ! {check_decoder_bin}"
    api = "D3D12"
    hw_acceleration = True


@register_decoder
class GStreamerLibvpxVP9(GStreamer10Video):
    """GStreamer VP9 Libvpx decoder implementation for GStreamer 1.0"""

    codec = Codec.VP9
    check_decoder_bin = " vp9dec  "
    decoder_bin = f" parsebin ! {check_decoder_bin}"
    api = "libvpx"
    hw_acceleration = False


@register_decoder
class GStreamerVaapiVP9Gst10Decoder(GStreamer10Video):
    """GStreamer VP9 VAAPI decoder implementation for GStreamer 1.0"""

    codec = Codec.VP9
    check_decoder_bin = " vaapivp9dec "
    decoder_bin = f" parsebin ! {check_decoder_bin}"
    api = "VAAPI"
    hw_acceleration = True


@register_decoder
class GStreamerVaapiAV1Gst10Decoder(GStreamer10Video):
    """GStreamer AV1 VAAPI decoder implementation for GStreamer 1.0"""

    codec = Codec.AV1
    check_decoder_bin = " vaapiav1dec "
    decoder_bin = f" parsebin ! {check_decoder_bin}"
    api = "VAAPI"
    hw_acceleration = True


@register_decoder
class GStreamerDav1dAV1Decoder(GStreamer10Video):
    """GStreamer AV1 dav1d decoder implementation for GStreamer 1.0"""

    codec = Codec.AV1
    check_decoder_bin = " dav1ddec "
    decoder_bin = f" parsebin ! {check_decoder_bin} "
    api = "dav1d"
    hw_acceleration = False


@register_decoder
class GStreamerVaVP9Gst10Decoder(GStreamer10Video):
    """GStreamer VP9 VA decoder implementation for GStreamer 1.0"""

    codec = Codec.VP9
    check_decoder_bin = " vavp9dec "
    decoder_bin = f" parsebin ! {check_decoder_bin}"
    api = "VA"
    hw_acceleration = True


@register_decoder
class GStreamerD3d11VP9Gst10Decoder(GStreamer10Video):
    """GStreamer VP9 D3D11 decoder implementation for GStreamer 1.0"""

    codec = Codec.VP9
    check_decoder_bin = " d3d11vp9dec "
    decoder_bin = f" parsebin ! {check_decoder_bin}"
    api = "D3D11"
    hw_acceleration = True


@register_decoder
class GStreamerD3d12VP9Gst10Decoder(GStreamer10Video):
    """GStreamer VP9 D3D12 decoder implementation for GStreamer 1.0"""

    codec = Codec.VP9
    check_decoder_bin = " d3d12vp9dec "
    decoder_bin = f" parsebin ! {check_decoder_bin}"
    api = "D3D12"
    hw_acceleration = True


@register_decoder
class GStreamerNvdecVP9Gst10Decoder(GStreamer10Video):
    """GStreamer VP9 NVDEC decoder implementation for GStreamer 1.0"""

    codec = Codec.VP9
    check_decoder_bin = " nvvp9dec "
    decoder_bin = f" parsebin ! {check_decoder_bin}"
    api = "NVDEC"
    hw_acceleration = True


@register_decoder
class GStreamerNvdecSLVP9Gst10Decoder(GStreamer10Video):
    """GStreamer VP9 NVDEC stateless decoder implementation for GStreamer 1.0"""

    codec = Codec.VP9
    check_decoder_bin = " nvvp9sldec "
    decoder_bin = f" parsebin ! {check_decoder_bin}"
    api = "NVDECSL"
    hw_acceleration = True


@register_decoder
class GStreamerVVdeCH266Decoder(GStreamer10Video):
    """GStreamer H.266/VVC VVdeC decoder implementation for GStreamer 1.0"""

    codec = Codec.H266
    decoder_bin = " h266parse ! vvdec "
    api = "VVdeC"
    hw_acceleration = False


@register_decoder
class FluendoH265Gst10Decoder(GStreamer10Video):
    """Fluendo H.265 software decoder implementation for GStreamer 1.0"""

    codec = Codec.H265
    decoder_bin = " h265parse ! fluh265dec "
    provider = "Fluendo"
    api = "SW"


@register_decoder
class FluendoH265Gst010Decoder(GStreamer010Video):
    """Fluendo H.265 software decoder implementation for GStreamer 0.10"""

    codec = Codec.H265
    decoder_bin = " h265parse ! fluh265dec "
    provider = "Fluendo"
    api = "SW"


@register_decoder
class FluendoH264Gst10Decoder(GStreamer10Video):
    """Fluendo H.264 software decoder implementation for GStreamer 1.0"""

    codec = Codec.H264
    decoder_bin = " h264parse ! fluh264dec "
    provider = "Fluendo"
    api = "SW"


@register_decoder
class FluendoH264Gst010Decoder(GStreamer010Video):
    """Fluendo H.264 software decoder implementation for GStreamer 0.10"""

    codec = Codec.H264
    decoder_bin = " fluh264dec "
    provider = "Fluendo"
    api = "SW"


@register_decoder
class FluendoH264VAGst10Decoder(GStreamer10Video):
    """Fluendo H.264 hardware decoder implementation for GStreamer 1.0"""

    codec = Codec.H264
    decoder_bin = " h264parse ! fluhwvadec "
    provider = "Fluendo"
    api = "HW"
    hw_acceleration = True


class FluendoH265VAGst10DecoderBase(GStreamer10Video):
    """Fluendo H.265 hardware decoder implementation for GStreamer 1.0"""

    codec = Codec.H265
    decoder_bin_tmpl = (
        " h265parse !"
        " video/x-h265,stream-format={stream_format},alignment={alignment} !"
        " fluhwvadec "
    )
    provider = "Fluendo"
    api = "HW"
    hw_acceleration = True
    stream_format: str = ""
    alignment: str = ""

    def __init__(self) -> None:
        super().__init__()
        self.name = self._translator(self.name)
        self.description = self._translator(self.description)
        self.decoder_bin = self.decoder_bin_tmpl.format(
            stream_format=self.stream_format, alignment=self.alignment
        )

    def _translator(self, target_val: str) -> str:
        new_val = f"{self.codec.value}-{self.stream_format}-{self.alignment}"
        return target_val.replace(self.codec.value, new_val)


@register_decoder
class FluendoH265ByteStreamAuVAGst10Decoder(FluendoH265VAGst10DecoderBase):
    """
    Fluendo H.265 (byte-stream/au) hardware decoder implementation for GStreamer 1.0
    """

    stream_format = "byte-stream"
    alignment = "au"


@register_decoder
class FluendoH265ByteStreamNalVAGst10Decoder(FluendoH265VAGst10DecoderBase):
    """
    Fluendo H.265 (byte-stream/nal) hardware decoder implementation for GStreamer 1.0
    """

    stream_format = "byte-stream"
    alignment = "nal"


@register_decoder
class FluendoH265Hev1AuVAGst10Decoder(FluendoH265VAGst10DecoderBase):
    """
    Fluendo H.265 (hev1/au) hardware decoder implementation for GStreamer 1.0
    """

    stream_format = "hev1"
    alignment = "au"


@register_decoder
class FluendoH265Hvc1AuVAGst10Decoder(FluendoH265VAGst10DecoderBase):
    """
    Fluendo H.265 (hvc1/au) hardware decoder implementation for GStreamer 1.0
    """

    stream_format = "hvc1"
    alignment = "au"


@register_decoder
class FluendoH265Hev1NalVAGst10Decoder(FluendoH265VAGst10DecoderBase):
    """
    Fluendo H.265 (hev1/nal) hardware decoder implementation for GStreamer 1.0
    """

    stream_format = "hev1"
    alignment = "nal"


@register_decoder
class FluendoH265Hvc1NalVAGst10Decoder(FluendoH265VAGst10DecoderBase):
    """
    Fluendo H.265 (hvc1/nal) hardware decoder implementation for GStreamer 1.0
    """

    stream_format = "hvc1"
    alignment = "nal"


@register_decoder
class FluendoFluVAH265DecGst10Decoder(GStreamer10Video):
    """Fluendo H.265 separated plugin hardware decoder for GStreamer 1.0"""

    codec = Codec.H265
    decoder_bin = " h265parse ! fluhwvah265dec "
    provider = "Fluendo"
    api = "HW"
    hw_acceleration = True
    name = f"{provider}-{codec.value}-{api}-hwvah265dec-Gst1.0"


@register_decoder
class FluendoFluVAH264DecGst10Decoder(GStreamer10Video):
    """Fluendo H.264 separated plugin hardware decoder for GStreamer 1.0"""

    codec = Codec.H264
    decoder_bin = " h264parse ! fluhwvah264dec "
    provider = "Fluendo"
    api = "HW"
    hw_acceleration = True
    name = f"{provider}-{codec.value}-{api}-hwvah264dec-Gst1.0"


@register_decoder
class FluendoFluAACDecGst10Decoder(GStreamer10Audio):
    """Fluendo AAC plugin decoder for GStreamer 1.0"""

    def __init__(self) -> None:
        self.codec = Codec.AAC
        self.decoder_bin = "fluaacdec trim=0"
        self.provider = "Fluendo"
        self.api = "SW"
        self.caps = self.caps + ",format=S16LE"
        super().__init__()



# Fluster - testing framework for decoders conformance
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <https://www.gnu.org/licenses/>.

import os
import argparse
import xml.etree.ElementTree as ET
import sys
import multiprocessing
import urllib.request

# pylint: disable=wrong-import-position
sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
from fluster.test_suite import TestSuite, TestVector
from fluster.codec import Codec, OutputFormat
from fluster import utils
from fluster.decoders import av1_aom

# pylint: enable=wrong-import-position

# Sourced from test/test_data_download_worker.cmake
AV1_URL = "https://storage.googleapis.com/aom-test-data"

# Sourced from test/test_vector_test.cc
BITSTREAM_EXTS = (
    ".ivf",
    ".webm",
    ".mkv"
)


class AOMGenerator:
    """Generates a test suite from the conformance bitstreams"""

    def __init__(
        self,
        name: str,
        suite_name: str,
        codec: Codec,
        description: str,
        site: str,
    ):
        self.name = name
        self.suite_name = suite_name
        self.codec = codec
        self.description = description
        self.site = site
        self.decoder = av1_aom.AV1AOMDecoder()

    def generate(self, download, jobs):
        """Generates the test suite and saves it to a file"""
        output_filepath = os.path.join(self.suite_name + ".json")
        test_suite = TestSuite(
            output_filepath,
            "resources",
            self.suite_name,
            self.codec,
            self.description,
            dict(),
        )

        print(f"Download list of bitstreams from {self.site + self.name}")
        opener = urllib.request.build_opener()
        root = ET.parse(opener.open(self.site)).getroot()
        opener.close()

        for entry in root.findall("{*}Contents"):
            if "Contents" not in entry.tag:
                continue
            fname = entry.find("{*}Key").text
            name = fname[:-4]
            if fname[-4:] not in BITSTREAM_EXTS or "invalid" in fname:
                continue

            file_url = f"{AV1_URL}/{entry.find('{*}Key').text}"
            test_vector = TestVector(
                name, file_url, "__skip__", fname, OutputFormat.YUV420P, "")
            test_suite.test_vectors[name] = test_vector

        if download:
            test_suite.download(
                jobs=jobs,
                out_dir=test_suite.resources_dir,
                verify=False,
                extract_all=True,
                keep_file=True,
            )

        for test_vector in test_suite.test_vectors.values():
            dest_dir = os.path.join(
                test_suite.resources_dir, test_suite.name, test_vector.name
            )
            dest_path = os.path.join(
                dest_dir, os.path.basename(test_vector.source))
            test_vector.input_file = dest_path.replace(
                os.path.join(
                    test_suite.resources_dir, test_suite.name, test_vector.name
                )
                + os.sep,
                "",
            )
            if not test_vector.input_file:
                raise Exception(f"Bitstream file not found in {dest_dir}")
            test_vector.source_checksum = utils.file_checksum(dest_path)
            out420 = f"{dest_path}.i420"
            # Run the libaom av1 decoder to get the checksum as the .md5 files are per-frame
            test_vector.result = self.decoder.decode(
                dest_path, out420, test_vector.output_format, 30, False)
            os.remove(out420)

        test_suite.to_json_file(output_filepath)
        print("Generate new test suite: " + test_suite.name + ".json")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--skip-download",
        help="skip extracting tarball",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "-j",
        "--jobs",
        help="number of parallel jobs to use. 2x logical cores by default",
        type=int,
        default=2 * multiprocessing.cpu_count(),
    )
    args = parser.parse_args()
    generator = AOMGenerator(
        "libaom-AV1",
        "AV1-TEST-VECTORS",
        Codec.AV1,
        "AV1 Test Vector Catalogue from https://storage.googleapis.com/aom-test-data",
        AV1_URL,
    )
    generator.generate(not args.skip_download, args.jobs)
